/* -------------------------------------------------------------------------- */
/* -                         Pyxis Rotator Control                          - */
/* -------------------------------------------------------------------------- */
/*                                                                            */
/* Copyright 2011 John Kielkopf                                               */
/*                                                                            */
/* Distributed under the terms of the General Public License (see LICENSE)    */
/*                                                                            */
/* John Kielkopf (kielkopf@louisville.edu)                                    */
/*                                                                            */
/* Date: July 25, 2011                                                        */
/* Version: 1.1                                                               */
/*                                                                            */
/* History:                                                                   */
/*                                                                            */
/* November 27, 2010                                                          */
/*   Version 1.0                                                              */
/*     Released                                                               */
/*                                                                            */
/* July 28, 2011                                                              */
/*   Version 1.1                                                              */
/*     Modified handling of non-responsive rotator                            */
/*     Increased setting speed for PA goto                                    */



#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <time.h>
#include <sys/time.h>
#include <unistd.h>
#include <fcntl.h>
#include <termios.h>

#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/types.h>

#include "rotate.h"


#define NULL_PTR(x) (x *)0

#ifndef TRUE
#define TRUE 1
#endif

#ifndef FALSE
#define FALSE 0
#endif


/* Communications variables and routines for internal use */

static int rotatorportfd;
static int rotatorconnectflag = FALSE;

/*  Commands */

int CheckConnectRotator(void);
int ConnectRotator(void);
int SetRotator(int rotatecmd, double angle);
int SetRotatorAngle(double angle);
int GetRotatorAngle(double *angle);
int WriteRotator(double angle);
int SetRotatorHome(void);
int DisconnectRotator(void);

/* Utility function */

double Map360(double angle);

/* Serial port utilities */

typedef fd_set telfds;

static int readn(int fd, char *ptr, int nbytes, int sec);
static int writen(int fd, char *ptr, int nbytes);
static int telstat(int fd,int sec,int usec);


/* Main program */

int main(int argc, char *argv[])
{
  double angle;
  int i;
  int rotatecmd; 
  int delay;
  char *testrotate;

  if (argc != 3) 
  { 
    printf("\n");
    printf("Usage: setrotate cmd angle\n");
    printf("\n");
    printf("Set the rotator and report current angle \n");
    printf("\n");
    printf("Command:  3 (sidereal)\n");
    printf("Command:  2 (delta)\n");
    printf("Command:  1 (goto)\n");
    printf("Command:  0 (home then goto)\n"); 
    printf("\n");
    return(0);
  }
  
  
  rotatecmd = strtod(argv[1],&testrotate);
  if ( (rotatecmd < 0) || (rotatecmd > 3) )
  {
    return(0);
  }
  
  angle = strtod(argv[2],&testrotate);
  
  if (rotatecmd != 2)
  {
    if ( (angle < 0.) || (angle > 359.999) )
    {
      return(0);
    }
  }
         
  i = 1;
  for ( ; ; )
  {
    if ( ConnectRotator() == TRUE )
    {
      break;
    }
    else if ( i > 10 )
    {     
      fprintf(stderr,"Cannot connect to rotator ...\n");
      return(0);
    }   
    else
    {
      i++;
    }
  }      
  SetRotator(rotatecmd,angle);
  usleep(1000000);
  GetRotatorAngle(&angle);
  WriteRotator(angle);
  DisconnectRotator();
  return(0);
} 

/* Report on rotator connection status */

int CheckConnectRotator(void)
{
  if (rotatorconnectflag == TRUE)
  {
    return(TRUE);
  }
  else
  {
    return(FALSE);
  }
}


/* Connect to the rotator serial interface */
/* Sets rotatorconnectflag TRUE and returns TRUE on success */

int ConnectRotator(void)
{
  struct termios tty;
  char sendstr[32]; 
  char returnstr[32];
  char rotatorport[32];
  int numread;
  
  DisconnectRotator();
  rotatorconnectflag = FALSE;
  
  /* Make the connection */

  strcpy(rotatorport,ROTATORPORT);
  rotatorportfd = open(rotatorport,O_RDWR);
  if(rotatorportfd == -1)
  {
    fprintf(stderr,"The rotator serial port is not available ... \n");
    return(0) ;
  }
    
  tcgetattr(rotatorportfd,&tty);
  cfsetospeed(&tty, (speed_t) B19200);
  cfsetispeed(&tty, (speed_t) B19200);
  tty.c_cflag = (tty.c_cflag & ~CSIZE) | CS8;
  tty.c_iflag =  IGNBRK;
  tty.c_lflag = 0;
  tty.c_oflag = 0;
  tty.c_cflag |= CLOCAL | CREAD;
  tty.c_cc[VMIN] = 1;
  tty.c_cc[VTIME] = 5;
  tty.c_iflag &= ~(IXON|IXOFF|IXANY);
  tty.c_cflag &= ~(PARENB | PARODD);
  tcsetattr(rotatorportfd, TCSANOW, &tty);

  /* Flush the input (read) buffer */

  tcflush(rotatorportfd,TCIOFLUSH);
  
  /* Establish communications */
  
  strcpy(sendstr,"CCLINK");
  writen(rotatorportfd,sendstr,6);
  numread=readn(rotatorportfd,returnstr,3,1);
    
  if ( returnstr[0]!='!' )
  {
    fprintf(stderr,"The rotator is responding ... \n"); 
    tcflush(rotatorportfd,TCIOFLUSH);
    DisconnectRotator();
    return(0);
  }
             
  rotatorconnectflag = TRUE;
      
  /* Flush the input buffer in case there is something left from startup */

  tcflush(rotatorportfd,TCIOFLUSH);

  return(1);
}
/* Home */

int SetRotatorHome(void)
{
  char sendstr[32];
  char returnstr[32];
  int flag;
  
  sprintf(sendstr,"CHOMES");
  writen(rotatorportfd,sendstr,6);
    
  /* Wait for 'F' to indicate completion of request */

  for (;;) 
  {
    flag = readn(rotatorportfd,returnstr,1,1);
    if (returnstr[0] == 'F') 
    {
      return(1);
    }  
    else if (returnstr[0] != '!')
    {
      return(0);
    }   
  }     

  return(0);
}

int SetRotator(int rotatecmd, double angle)
{
  double nowangle;
  int speed; 
    
  /* Set speed fast then home and go to angle */
  if (rotatecmd == 0)
  {
    speed = 2;
    SetRotatorSpeed();    
    fprintf(stderr,"Homing rotator ...\n",angle);
    SetRotatorHome();
    usleep(1000000);
    SetRotatorAngle(angle);
    return(1);
  }    
  
  /* Set speed fast then go to angle */
  if (rotatecmd == 1)  
  {    
    speed = 2;
    SetRotatorSpeed(speed); 
    usleep(1000000);
    SetRotatorAngle(angle);
    return(1);
  }
  
  /* Set speed fast and go to delta  */
  if (rotatecmd == 2)
  {    
    speed = 2;
    SetRotatorSpeed(speed); 
    usleep(1000000);
    GetRotatorAngle(&nowangle);
    usleep(1000000);
    angle = angle + nowangle;
    angle = Map360(angle);
    SetRotatorAngle(angle);
    return(1);
  }
  
  /* Run at sidereal rate */
   
  if (rotatecmd == 3)
  {
    fprintf(stderr,"Support for sidereal rate on Pyxis not yet implemented.\n");
    return(0);
  }
  
  fprintf(stderr,"Unknown rotate command.\n");
  return(0);
}
    

/* Set rotator to selected angle */

int SetRotatorAngle(double angle)
{
  char sendstr[32];
  char returnstr[32];
  int rotatorangle = 0;
  int flag;

  if ((angle < 0.) || (angle > 359.999))
  {
    fprintf(stderr,"Position angle must be in the range 0 to 359.999 degrees.\n");
    return(0);
  }
  
  /* Pyxis uses integer angles */
  
  sprintf(sendstr,"CPA000");
  rotatorangle = (int) angle;
  
  /* The new position angle is entered as nnn degrees*/
  /* Pyxis accepts integer values from 000 to 359 */
  /* It requires left justified zeros as fill: 181 081 or 001 */
    
  sprintf(sendstr,"CPA%3d",rotatorangle);
  if (rotatorangle < 100)
  {
    sendstr[3] = '0';
  }
  if (angle < 10.)
  {
    sendstr[4] = '0';
  }  

  flag = writen(rotatorportfd,sendstr,6);
  
  /* Wait for 'F' to indicate completion of request */
  /* Each step returns a ! which we ignore here */

  for (;;) 
  {
    flag = readn(rotatorportfd,returnstr,1,1);
    returnstr[1]='\0';
    if (returnstr[0] == 'F') 
    {
    
      /* Flush the buffer */
      
      tcflush(rotatorportfd,TCIOFLUSH);

      return(1);
    }  
    else if (returnstr[0] == '2')
    {
      
      /* Error code 2 for requesting rotation to current position */      
      /* Flush the buffer */
      
      tcflush(rotatorportfd,TCIOFLUSH);

      return(0);
    }   
  }   

}


/* Set rotator to selected speed*/

int SetRotatorSpeed(int speed)
{
  char sendstr[32];
  char returnstr[32];
  int flag;

  if ((speed < 0) || (speed > 99))
  {
    fprintf(stderr,"Speed must be in the range 0 to 99.\n");
    return(0);
  }
  
  /* Pyxis uses integer angles */
  
  /* Set the default rate */
  
  sprintf(sendstr,"CT0008");
  
  /* The new speed is entered as nn units */
  /* Pyxis accepts integer values from 00 to 99 */
  /* It requires left justified zeros as fill: 99 or 09 */
    
  sprintf(sendstr,"CT00%2d",speed);
  if (speed < 10)
  {
    sendstr[4] = '0';
  }

  flag = writen(rotatorportfd,sendstr,6);
  tcflush(rotatorportfd,TCIOFLUSH);
  return(1);

}


/* Return the current rotator angle */

int GetRotatorAngle(double *angle)
{
  char sendstr[32];
  char returnstr[32];
  int numread;
  int rotator_angle;

  /* Flush the input (read) buffer */

  tcflush(rotatorportfd,TCIOFLUSH);

  sprintf(sendstr,"CGETPA");
  writen(rotatorportfd,sendstr,6);
  numread=readn(rotatorportfd,returnstr,5,1);
  returnstr[3] = '\0'; 
  sscanf(returnstr,"%d",&rotator_angle);
  *angle = rotator_angle;   
  return(1);
}


/* Close serial connection to  IFW and reset rotatorconnectflag */

int DisconnectRotator(void)
{
  if(rotatorconnectflag == TRUE)
  {
    close(rotatorportfd);
    rotatorconnectflag = FALSE;
    return(TRUE);
  }
  rotatorconnectflag = FALSE;
  return(FALSE);
}


/* Write rotator reading to stdout and to a system status file */

int WriteRotator(double angle)
{
  FILE* outfile;
  outfile = fopen("/usr/local/observatory/status/telrotate","w");
  if ( outfile == NULL )
  {
    fprintf(stderr,"Cannot update telrotate status file\n");
    return FALSE;
  }

  fprintf(stdout, "%.1f\n", angle); 
  fprintf(outfile, "%lf\n", angle);      
  fclose(outfile);
  return TRUE;
}

/* Map an angle in degrees to  0 <= angle < 360 */

double Map360(double angle)
{
  int n;

  if (angle < 0.0) 
  {
    n = (int) (angle / 360.0) - 1;
   return (angle - n * 360.0);
  } 
  else if (angle >= 360.0)
  {
    n = (int) (angle / 360.0);
    return (angle - n * 360.0);
  } 
  else 
  {
    return (angle);
  }
}



/* Serial port utilities */

static int writen(fd, ptr, nbytes)
int fd;
char *ptr;
int nbytes;
{
  int nleft, nwritten;
  nleft = nbytes;
  while (nleft > 0) 
  {
    nwritten = write (fd, ptr, nleft);
    if (nwritten <=0 ) break;
    nleft -= nwritten;
    ptr += nwritten;
  }
  return (nbytes - nleft);
}

static int readn(fd, ptr, nbytes, sec)
int fd;
char *ptr;
int nbytes;
int sec;
{
  int stat;
  int nleft, nread;
  nleft = nbytes;
  while (nleft > 0) 
  {
    stat = telstat(fd,sec,0);
    if (stat <=  0 ) break;
    nread  = read (fd, ptr, nleft);

/*  Diagnostic */    
    
/*    printf("readn: %d read\n", nread);  */

    if (nread <= 0)  break;
    nleft -= nread;
    ptr += nread;
  }
  return (nbytes - nleft);
}

/*
 * Examines the read status of a file descriptor.
 * The timeout (sec, usec) specifies a maximum interval to
 * wait for data to be available in the descriptor.
 * To effect a poll, the timeout (sec, usec) should be 0.
 * Returns non-negative value on data available.
 * 0 indicates that the time limit referred by timeout expired.
 * On failure, it returns -1 and errno is set to indicate the
 * error.
 */
static int telstat(fd,sec,usec)
register int fd, sec, usec;
{
  int ret;
  int width;
  struct timeval timeout;
  telfds readfds;

  memset((char *)&readfds,0,sizeof(readfds));
  FD_SET(fd, &readfds);
  width = fd+1;
  timeout.tv_sec = sec;
  timeout.tv_usec = usec;
  ret = select(width,&readfds,NULL_PTR(telfds),NULL_PTR(telfds),&timeout);
  return(ret);
}

