/* -------------------------------------------------------------------------- */
/* -                   Astronomical Telescope Control                       - */
/* -                        XmTel User Interface                            - */
/* -                                                                        - */
/* -                          XPA for XmTel                                 - */
/* -                                                                        - */
/* -------------------------------------------------------------------------- */
/*                                                                            */
/* Copyright (c) 2008-2014 John Kielkopf                                      */
/* kielkopf@louisville.edu                                                    */
/*                                                                            */
/* This file is part of XmTel.                                                */
/*                                                                            */
/* Distributed under the terms of the MIT License (see LICENSE)               */
/*                                                                            */
/* Date: November 1, 2014                                                     */
/* Version: 7.0                                                               */
/*                                                                            */
/* About epochs ...                                                           */
/*                                                                            */
/*   XmTel runs on Epoch of Date (EOD) coordinates                            */
/*   The user interface accepts Epoch JD 2000.0 (J2000) coordinates           */
/*                                                                            */
/* See TELCOMMANDS for the xpa commands used by xmtel                         */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <fcntl.h>
#include <unistd.h>
#include <time.h>
#include <sys/time.h>
#include <sys/types.h>
#include <ctype.h>

#include <Xm/Xm.h>
#include <Xm/Form.h>
#include <Xm/ToggleB.h>
#include <Xm/FileSB.h>
#include <Xm/PushB.h>
#include <Xm/Label.h>
#include <Xm/RowColumn.h> 
#include <Xm/CascadeB.h>
#include <Xm/MessageB.h>
#include <Xm/Separator.h>
#include <Xm/SelectioB.h>
#include <Xm/Text.h>
#include <Xm/List.h>

#include <xpa.h>

#include "xmtel.h"

XtAppContext context;
XmStringCharSet char_set=XmSTRING_DEFAULT_CHARSET;

/* Widgets for the menu bar */

Widget menu_bar;
Widget file_menu;
Widget newqueue_item;
Widget newlog_item;
Widget exit_item;
Widget edit_menu;
Widget editqueue_item;
Widget editlog_item;
Widget pointing_menu;
Widget p_options_toggle_1;
Widget p_options_toggle_2;
Widget p_options_toggle_4;
Widget p_options_toggle_8;
Widget p_options_toggle_16;
Widget p_options_toggle_32;
Widget ref_menu;
Widget ref_target_item;
Widget ref_wcs_item;
Widget ref_clear_item;
Widget ref_save_item;
Widget ref_recall_item;
Widget model_menu;
Widget model_edit_item;
Widget model_clear_item;
Widget model_save_item;
Widget model_recall_item;
Widget model_default_item;


/* Widgets for the panel */

Widget toplevel;
Widget dialog;
Widget button[3][3];
Widget command[12];
Widget toggle[4];
Widget form;
Widget sep_btns;
Widget sep_coords;
Widget sep_queue;
Widget sep_status;
Widget label_mark;
Widget label_ra;
Widget label_dec;
Widget label_epoch;
Widget mark_tel;
Widget mark_target;

Widget get_telra;
Widget get_teldec;
Widget get_telepoch;
Widget get_targetra;
Widget get_targetdec;
Widget get_targetepoch;

Widget select_logfile;
Widget read_queuefile;

Widget queue_area;
Widget message_area;
Widget radio_box;

/* User interface polling */

unsigned long poll_interval = POLLMS;      
void poll_interval_handler(XtPointer client_data_ptr, XtIntervalId *client_id);
XtIntervalId poll_interval_id;
XtPointer poll_interval_data_ptr;

Widget make_menu_item();     /* add a pushbutton item to the menu */
Widget make_menu_toggle();   /* add a toggle item to the menu     */
Widget make_menu();          /* create a menu on the menu bar     */

/* Widgets for the model_edit panel */

Widget model_edit;


/* Labels for the guide buttons top to bottom then left to right */

static char btn_name[][6] = {
    "NoOp", 
    "E",
    "NoOp",
    "N",
    "NoOp",
    "S",
    "NoOp",
    "W",
    "NoOp"
};

/* Simple labels for the command area top to bottom, left to right */

static char cmd_name[][15] = {
    "Slew",
    "Save",
    "Track",
    "Recall",
    "Guide",  
    "Home",  
    "Stop",
    "Park"
};

/* Labels for the guide speed settings */

static char spd_name[][8] = {
    "Slew",
    "Find",
    "Center",
    "Guide"   
};


/* Message area string */

char message[300];


/* Menu functions */

void create_menus(Widget menu_bar);  /* Creates all menus for program         */
void menuCB();             /* Callback routine used for all menus             */
void setup_model_edit();   /* Create the unmanaged model edit panel           */
void model_editCB();       /* Callback function for model edit panel          */

void dialogCB();           /* Callback function for the dialog box            */
void buttonCB();           /* Callback function for the paddle buttons        */
void speedCB();            /* Callback function for the slew speed buttons    */
void teldataCB();          /* Callback function for the data buttons          */
void commandCB();          /* Callback function for the command buttons       */
void selectlogCB();        /* Callback function for selecting the log file    */
void readqueueCB();        /* Callback function for opening the queue file    */
void selectqueueCB();      /* Callback function for selecting a queue entry   */

/* Message display functions */

void show_telescope_coordinates();      /* Display latest telescope coordinates */
void show_target_coordinates();         /* Display latest target coordinates  */
void show_model_parameters();           /* Display latest model parameters */
void show_guide_status();               /* Display the guide status */

/* Message area display */

void show_message();                    /* Update message area with string message */

/* Local log and queue operations */

void save_coordinates();                /* Save a log entry and add to the history */
void recall_coordinates();              /* Read the next previous history entry */
void read_queue();                      /* Read queue file into memory */

/* User interface RA and Dec direct entry */

void new_target_ra();                   /* User input of target RA */
void new_target_dec();                  /* User input of target Dec */

/* Interface to a local sky display */

void link_fifos();                      /* Startup fifo link routine */
void unlink_fifos();                    /* Shutdown fifo link routine */
void mark_xephem_telescope();           /* Export telescope coordinates to xephem */
void read_xephem_target();              /* Read target coordinates from xephem goto fifo */
void mark_xephem_target();              /* Wxport target coordinates to xephem */

/* Interface to xpa */

void tel_send_telescope(void);
void tel_receive_telescope(void);
void tel_send_target(void);
void tel_receive_target(void);
void tel_send_status(void);
void tel_receive_status(void);
void tel_send_track(void);
void tel_receive_track(void);
void tel_send_goto(void);
void tel_receive_goto(void);
void tel_send_guide(void);
void tel_receive_guide(void);
void tel_send_speed(void);
void tel_receive_speed(void);
void tel_send_move(void);
void tel_receive_move(void);
void tel_send_refract(void);
void tel_receive_refract(void);
void tel_send_dynamic(void);
void tel_receive_dynamic(void);
void tel_send_offset(void);
void tel_send_offset_target(void);
void tel_send_offset_clear(void);
void tel_receive_offset(void);
void tel_send_polar(void);
void tel_receive_polar(void);
void tel_send_park(void);
void tel_receive_park(void);
void tel_send_home(void);
void tel_receive_home(void);

/* Convert string deg:min:sec to double */

int dmstod (char *instr, double *datap);

/* Convert dms double to string deg:min:sec */

void dtodms (char *outstr, double *dmsp); 

/* Site Parameters (Actual values are defined in the driver ) */

double SiteLatitude = 85.5300;       /* Latitude in degrees + north */  
double SiteLongitude = 38.3334;      /* Longitude in degrees + west */  
double SiteAltitude = 230.0;         /* Altitude in meters */
double SitePressure = 760.0;         /* Atmospheric pressure in Torr */
double SiteTemperature = 20.0;       /* Local atmospheric temperature  C */

/* Telescope control */

char   telescope_device[64];


/* Telescope parameters used by this client */

double telra;                 /* Epoch of date!!          */
double teldec;                /* Epoch of date!!          */
double telra2;                /* Epoch 2000.0             */
double teldec2;               /* Epoch 2000.0             */  
double targetra;              /* Epoch of date!!          */
double targetdec;             /* Epoch of date!!          */
double targetra2;             /* Epoch 2000.0             */
double targetdec2;            /* Epoch 2000.0             */ 
double offsetha;
double offsetdec; 
double movetime;

/* Model parameters used by this client */

double modelha1;              /* Pixels/hour HA drift */
double modeldec1;             /* Pixels/hour dec drift */
double polaraz;
double polaralt;


/* User interface flags and variables  */

int quiet = TRUE;                  /* set FALSE for diagnostics */
int telinitflag=FALSE;                /* User interface ready flag */
int display_telepoch=EOD;             /* Display epoch for telescope */
int display_targetepoch=EOD;          /* Display epoch for target  */
int pmodel=RAW;                       /* Pointing model default to raw data */
int telspd=FIND;                      /* Drive speed */
int telflag=STATUSOFF;                /* Telescope connection flag */
int target=RA;                        /* Flag for target field input */
int trackflag=FALSE;                  /* Flag to indicate tracking state */
int guideflag=FALSE;                  /* Flag to indicate guiding state */
int teldir=WEST;                      /* Flag to indicate current slew direction */
int gotoflag=FALSE;                   /* Flag to indicate status of a goto request */
int parkflag=FALSE;                   /* Flag to indicate status of a goto park request */
int homeflag=FALSE;               /* Flag to indicate status of a goto home request */


/* Corrections for proper motion, precession, aberration, and nutation */

extern void Apparent(double *ra, double *dec, int dirflag);  
extern void PrecessToEOD(double epoch, double  *ra, double  *dec);
extern void PrecessToEpoch(double epoch, double  *ra, double  *dec);
extern void ProperMotion(double epoch, double *ra, double *dec, 
  double pm_ra, double pm_dec); 


/* Time */

extern double LSTNow(void);
extern double UTNow(void);


/* Utility */

extern double Map12(double ha);
extern double Map24(double ra);
extern double Map180(double dec);


/* Files */

int fd_fifo_in, fd_fifo_out;           /* FIFO file descriptors */
FILE *fp_log;                          /* Log file pointer      */
static char *logfile;                  /* Log name */
FILE *fp_queue;                        /* Queue file pointer */
static char *queuefile;                /* Queue name */


/* External commands */

static char *editor;                   /* External queue and log editor */


/* Buffers */

char buf[256];                         /* General purpose character buffer */


/* Queue */

int nqueue = 0;                        /* Number of entries in the queue */
int queuechoice = 0;                   /* Serial number of selected entry */


/* History */

/* The history catalog is a ring buffer starting at 0 for the first entry. */
/* There are nhistory entries in use (<=100). */

int nhistory = 0;                      /* Number of history entries (max 100) */
int last_history_saved = -1;           /* Location for last history saved */
int last_history_recalled = -1;        /* Location for last history recalled */

typedef struct                         /* Catalog structure assuming EOD */
{
  char name[80];
  char desc[80];
  double ra;
  double dec;
  double mag;
} catalog;
  
/* Create the space for catalogs.                                          */
/* Allow one extra entry and watch for maximum number in operations.       */
/* History is indexed from 1 for first entry, so 0th entry is not used.    */

catalog queue[10001];                  /* This holds the observing queue   */
catalog history[101];                  /* This holds the saved coordinates */


int main(argc,argv)
  int argc; 
  char *argv[];
{
  Arg al[10];
  int ac;
  unsigned long x,y,btn_number,cmd_number;
  EventMask mask;
  
  /* No command line arguments needed */
  
  sprintf(telescope_device,"tel");  

  /* User interface not initialized yet */
  
  telinitflag = FALSE;

  /* Make contact with the server */
  
  telflag = STATUSOFF;
  
  tel_receive_status();
  if (telflag == STATUSOFF)
  {
    fprintf(stderr,"XPA server is not available\n");
    exit(1);
  }

  
  fprintf(stderr,"XPA server connected\n");
  

  /* Define the defaults for log and queue files */
  
  logfile = (char *) malloc (MAXPATHLEN);
  strcpy(logfile,LOGFILE);
  queuefile = (char *) malloc (MAXPATHLEN);
  strcpy(queuefile,QUEUEFILE);
  editor = (char *) malloc (MAXPATHLEN);
  strcpy(editor,XMTEL_EDITOR);

  /* Start the fifo communications */
  /* Start XEphem after XmTel is running */
  
  link_fifos();
 
  /* Create the toplevel shell */

  toplevel = XtAppInitialize(&context,"",NULL,0,&argc,argv,
      NULL,NULL,0);

  /* Set the default size of the window */ 

  ac=0;
  XtSetArg(al[ac],XmNwidth,400); ac++;
  XtSetArg(al[ac],XmNheight,700); ac++;
  XtSetValues(toplevel,al,ac);

  /* Create a form widget */

  ac=0;
  form=XmCreateForm(toplevel,"form",al,ac);
  XtManageChild(form);

  /* Create the menu bar and attach it to the form */
  
  ac=0;
  XtSetArg(al[ac],XmNtopAttachment,XmATTACH_FORM); ac++;
  XtSetArg(al[ac],XmNrightAttachment,XmATTACH_FORM); ac++;
  XtSetArg(al[ac],XmNleftAttachment,XmATTACH_FORM); ac++;
  menu_bar=XmCreateMenuBar(form,"menu_bar",al,ac);
  XtManageChild(menu_bar);

  /* Note that locations of buttons are in percent of window dimension */

  /* Create a radio box to hold the speed toggles and attach it to form */
  
  ac=0;
  XtSetArg(al[ac],XmNorientation,XmVERTICAL); ac++;
  XtSetArg(al[ac],XmNtopAttachment,XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNtopPosition,12); ac++; 
  XtSetArg(al[ac],XmNleftAttachment,XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNleftPosition,5); ac++; 
  XtSetArg(al[ac],XmNrightAttachment,XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNrightPosition,24); ac++; 
  radio_box=XmCreateRadioBox(form,"radio_box",al,ac);
  XtManageChild(radio_box);
  
  /* Create speed toggles */
  
  for (x=0; x<4; x++)
  {
    ac=0;
    XtSetArg(al[ac],XmNlabelString,
      XmStringCreate(spd_name[x],char_set)); ac++;
    toggle[x]=XmCreateToggleButton(radio_box,"toggle",al,ac);
    XtManageChild(toggle[x]);
    XtAddCallback (toggle[x], XmNvalueChangedCallback, speedCB, (XtPointer) x);
  }     
 
  /* Create the slew speed togglebuttons with find speed selected */

  XmToggleButtonSetState (toggle[1], True, False);
  
  /* Set up the slew buttons, add event handlers, attach to form. */

  for (x=0; x<3; x++)
  {
    for (y=0; y<3; y++)
    {
      btn_number=3*x+y;
      if ( (btn_number==1) | (btn_number==3) | (btn_number==5) | (btn_number==7) ) 
      {
        ac=0;
        XtSetArg(al[ac],XmNlabelString,
            XmStringCreate(btn_name[btn_number],char_set)); ac++;
        XtSetArg(al[ac],XmNleftAttachment,
            XmATTACH_POSITION); ac++;
        XtSetArg(al[ac],XmNleftPosition,28+x*15); ac++;
        XtSetArg(al[ac],XmNrightAttachment,
            XmATTACH_POSITION); ac++;
        XtSetArg(al[ac],XmNrightPosition,42+x*15); ac++;
        XtSetArg(al[ac],XmNtopAttachment,
            XmATTACH_POSITION); ac++;
        XtSetArg(al[ac],XmNtopPosition,6+y*9); ac++;
        XtSetArg(al[ac],XmNbottomAttachment,
            XmATTACH_POSITION); ac++;
        XtSetArg(al[ac],XmNbottomPosition, 15+y*9); ac++;
        button[x][y]=XmCreatePushButton(form,"label",al,ac);
        mask = ButtonReleaseMask | ButtonPressMask;
        XtAddEventHandler(button[x][y],mask,False,
            buttonCB,(XtPointer) btn_number);              
        XtManageChild(button[x][y]);
      }
    }
  }

  /* Create a separator widget and attach it to the form */

  ac=0;
  XtSetArg(al[ac],XmNtopAttachment,XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNtopPosition,35); ac++;
  XtSetArg(al[ac],XmNrightAttachment,XmATTACH_FORM); ac++; 
  XtSetArg(al[ac],XmNleftAttachment,XmATTACH_FORM); ac++;  
  XtSetArg(al[ac],XmNbottomAttachment,XmATTACH_NONE); ac++;  
  sep_btns=XmCreateSeparator(form,"sep",al,ac);
  XtManageChild(sep_btns);   
 
  /* Build the telescope and target coordinates display area */
  
  /* Create the mark label widget and attach it to the form */

  ac=0;
  XtSetArg(al[ac],XmNlabelString,
      XmStringCreate("Mark",char_set)); ac++;
  XtSetArg(al[ac],XmNleftAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNleftPosition,13); ac++;
  XtSetArg(al[ac],XmNrightAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNrightPosition,31); ac++;
  XtSetArg(al[ac],XmNtopAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNtopPosition,36); ac++;
  XtSetArg(al[ac],XmNbottomAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNbottomPosition,40); ac++;
  label_mark=XmCreateLabel(form,"label",al,ac);
  XtManageChild(label_mark); 
  
  /* Create the ra label widget and attach it to the form */

  ac=0;
  XtSetArg(al[ac],XmNlabelString,
      XmStringCreate("RA",char_set)); ac++;
  XtSetArg(al[ac],XmNleftAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNleftPosition,37); ac++;
  XtSetArg(al[ac],XmNrightAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNrightPosition,45); ac++;
  XtSetArg(al[ac],XmNtopAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNtopPosition,36); ac++;
  XtSetArg(al[ac],XmNbottomAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNbottomPosition,40); ac++;
  label_ra=XmCreateLabel(form,"label",al,ac);
  XtManageChild(label_ra); 

  /* Create the dec label widget and attach it to the form */

  ac=0;
  XtSetArg(al[ac],XmNlabelString,
      XmStringCreate("Dec",char_set)); ac++;
  XtSetArg(al[ac],XmNleftAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNleftPosition,55); ac++;
  XtSetArg(al[ac],XmNrightAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNrightPosition,63); ac++;
  XtSetArg(al[ac],XmNtopAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNtopPosition,36); ac++;
  XtSetArg(al[ac],XmNbottomAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNbottomPosition,40); ac++;
  label_dec=XmCreateLabel(form,"label",al,ac);
  XtManageChild(label_dec);    

  /* Create the epoch label widget and attach it to the form */

  ac=0;
  XtSetArg(al[ac],XmNlabelString,
      XmStringCreate("Epoch",char_set)); ac++;
  XtSetArg(al[ac],XmNleftAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNleftPosition,72); ac++;
  XtSetArg(al[ac],XmNrightAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNrightPosition,80); ac++;
  XtSetArg(al[ac],XmNtopAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNtopPosition,36); ac++;
  XtSetArg(al[ac],XmNbottomAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNbottomPosition,40); ac++;
  label_epoch=XmCreateLabel(form,"label",al,ac);
  XtManageChild(label_epoch);    

  /* Create the telescope label widget and attach it to the form */

  cmd_number = 0;
  ac=0;
  XtSetArg(al[ac],XmNlabelString,
      XmStringCreate("Telescope",char_set)); ac++;
  XtSetArg(al[ac],XmNleftAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNleftPosition,13); ac++;
  XtSetArg(al[ac],XmNrightAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNrightPosition,31); ac++;
  XtSetArg(al[ac],XmNtopAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNtopPosition,40); ac++;
  XtSetArg(al[ac],XmNbottomAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNbottomPosition,44); ac++;
  mark_tel=XmCreatePushButton(form,"label",al,ac);
  XtManageChild(mark_tel);
  XtAddCallback(mark_tel,XmNactivateCallback,
                teldataCB,(XtPointer) cmd_number);     
       
  /* Create the target label widget and attach it to the form */

  cmd_number = 1;
  ac=0;
  XtSetArg(al[ac],XmNlabelString,
      XmStringCreate("Target",char_set)); ac++;
  XtSetArg(al[ac],XmNleftAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNleftPosition,13); ac++;
  XtSetArg(al[ac],XmNrightAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNrightPosition,31); ac++;
  XtSetArg(al[ac],XmNtopAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNtopPosition,44); ac++;
  XtSetArg(al[ac],XmNbottomAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNbottomPosition,48); ac++;
  mark_target=XmCreatePushButton(form,"label",al,ac);
  XtManageChild(mark_target);    
  XtAddCallback(mark_target,XmNactivateCallback,
                teldataCB,(XtPointer) cmd_number);     
 
  /* Create a telescope ra widget and attach it to the form */

  cmd_number = 2;    
  ac=0;
  XtSetArg(al[ac],XmNlabelString,
      XmStringCreate("00:00:00",char_set)); ac++;
  XtSetArg(al[ac],XmNleftAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNleftPosition,33); ac++;
  XtSetArg(al[ac],XmNrightAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNrightPosition,49); ac++;
  XtSetArg(al[ac],XmNtopAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNtopPosition,40); ac++;
  XtSetArg(al[ac],XmNbottomAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNbottomPosition,44); ac++;
  get_telra=XmCreatePushButton(form,"label",al,ac);
  XtManageChild(get_telra);
  XtAddCallback(get_telra,XmNactivateCallback,
                teldataCB,(XtPointer) cmd_number);     
 
  /* Create a telescope dec widget and attach it to the form */

  cmd_number = 3;    
  ac=0;
  XtSetArg(al[ac],XmNlabelString,
      XmStringCreate("00:00:00",char_set)); ac++;
  XtSetArg(al[ac],XmNleftAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNleftPosition,51); ac++;
  XtSetArg(al[ac],XmNrightAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNrightPosition,67); ac++;
  XtSetArg(al[ac],XmNtopAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNtopPosition,40); ac++;
  XtSetArg(al[ac],XmNbottomAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNbottomPosition,44); ac++;
  get_teldec=XmCreatePushButton(form,"label",al,ac);
  XtManageChild(get_teldec);
  XtAddCallback(get_teldec,XmNactivateCallback,
                teldataCB,(XtPointer) cmd_number);    

  /* Create a telescope epoch widget and attach it to the form */

  cmd_number = 4;  
  ac=0;
  XtSetArg(al[ac],XmNlabelString,
      XmStringCreate("Now",char_set)); ac++;
  XtSetArg(al[ac],XmNleftAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNleftPosition,69); ac++;
  XtSetArg(al[ac],XmNrightAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNrightPosition,83); ac++;
  XtSetArg(al[ac],XmNtopAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNtopPosition,40); ac++;
  XtSetArg(al[ac],XmNbottomAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNbottomPosition,44); ac++;
  get_telepoch=XmCreatePushButton(form,"label",al,ac);
  XtManageChild(get_telepoch);
  XtAddCallback(get_telepoch,XmNactivateCallback,
                teldataCB,(XtPointer) cmd_number);     
  
  
  /* Create a target ra widget and attach it to the form */

  cmd_number = 5;  
  ac=0;
  XtSetArg(al[ac],XmNlabelString,
      XmStringCreate("00:00:00",char_set)); ac++;
  XtSetArg(al[ac],XmNleftAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNleftPosition,33); ac++;
  XtSetArg(al[ac],XmNrightAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNrightPosition,49); ac++;
  XtSetArg(al[ac],XmNtopAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNtopPosition,44); ac++;
  XtSetArg(al[ac],XmNbottomAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNbottomPosition,48); ac++;
  get_targetra=XmCreatePushButton(form,"label",al,ac);
  XtManageChild(get_targetra);
  XtAddCallback(get_targetra,XmNactivateCallback,
                teldataCB,(int *) cmd_number);
                        

  /* Create a target dec widget and attach it to the form */

  cmd_number = 6;  
  ac=0;
  XtSetArg(al[ac],XmNlabelString,
      XmStringCreate("00:00:00",char_set)); ac++;
  XtSetArg(al[ac],XmNleftAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNleftPosition,51); ac++;
  XtSetArg(al[ac],XmNrightAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNrightPosition,67); ac++;
  XtSetArg(al[ac],XmNtopAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNtopPosition,44); ac++;
  XtSetArg(al[ac],XmNbottomAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNbottomPosition,48); ac++;
  get_targetdec=XmCreatePushButton(form,"label",al,ac);
  XtManageChild(get_targetdec);
  XtAddCallback(get_targetdec,XmNactivateCallback,
                teldataCB,(XtPointer) cmd_number);    

  
  /* Create a target epoch widget and attach it to the form */

  cmd_number = 7;  
  ac=0;
  XtSetArg(al[ac],XmNlabelString,
      XmStringCreate("Now",char_set)); ac++;
  XtSetArg(al[ac],XmNleftAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNleftPosition,69); ac++;
  XtSetArg(al[ac],XmNrightAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNrightPosition,83); ac++;
  XtSetArg(al[ac],XmNtopAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNtopPosition,44); ac++;
  XtSetArg(al[ac],XmNbottomAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNbottomPosition,48); ac++;
  get_targetepoch=XmCreatePushButton(form,"label",al,ac);
  XtManageChild(get_targetepoch);
  XtAddCallback(get_targetepoch,XmNactivateCallback,
                teldataCB,(XtPointer) cmd_number);        
  

  /* Create a separator widget and attach it to the form */

  ac=0;
  XtSetArg(al[ac],XmNtopAttachment,XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNtopPosition,50); ac++;
  XtSetArg(al[ac],XmNrightAttachment,XmATTACH_FORM); ac++; 
  XtSetArg(al[ac],XmNleftAttachment,XmATTACH_FORM); ac++;  
  XtSetArg(al[ac],XmNbottomAttachment,XmATTACH_NONE); ac++;      
  sep_coords=XmCreateSeparator(form,"sep",al,ac);
  XtManageChild(sep_coords);  
  

  /* Queue area */ 
  
  /* Create a scrolled queue list widget and attach it to the form */

  ac=0;
  XtSetArg(al[ac],XmNleftAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNleftPosition,8); ac++;
  XtSetArg(al[ac],XmNrightAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNrightPosition,92); ac++;
  XtSetArg(al[ac],XmNtopAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNtopPosition,52); ac++;
  XtSetArg(al[ac],XmNbottomAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNbottomPosition,63); ac++;
  queue_area=XmCreateScrolledList(form,"queue_area",al,ac);
  XtManageChild(queue_area);
  XtAddCallback(queue_area,XmNdefaultActionCallback,selectqueueCB,NULL);
  XtAddCallback(queue_area,XmNbrowseSelectionCallback,selectqueueCB,NULL);

    
  /* Create a separator widget and attach it to the form */

  ac=0;
  XtSetArg(al[ac],XmNtopAttachment,XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNtopPosition,65); ac++;
  XtSetArg(al[ac],XmNrightAttachment,XmATTACH_FORM); ac++; 
  XtSetArg(al[ac],XmNleftAttachment,XmATTACH_FORM); ac++;  
  XtSetArg(al[ac],XmNbottomAttachment,XmATTACH_NONE); ac++;      
  sep_queue=XmCreateSeparator(form,"sep",al,ac);
  XtManageChild(sep_queue);   

  /* Set up the 8 command buttons and attach them to the form */

  for (x=0; x<4; x++)   
  for (y=0; y<2; y++)
  {
    cmd_number=2*x+y;
    ac=0;
    XtSetArg(al[ac],XmNlabelString,
      XmStringCreate(cmd_name[cmd_number],char_set)); ac++;
    XtSetArg(al[ac],XmNleftAttachment,
      XmATTACH_POSITION); ac++;
    XtSetArg(al[ac],XmNleftPosition,3+x*24); ac++;
    XtSetArg(al[ac],XmNrightAttachment,
      XmATTACH_POSITION); ac++;
    XtSetArg(al[ac],XmNrightPosition,25+x*24); ac++;
    XtSetArg(al[ac],XmNtopAttachment,
      XmATTACH_POSITION); ac++;
    XtSetArg(al[ac],XmNtopPosition,67+y*9); ac++;
    XtSetArg(al[ac],XmNbottomAttachment,
      XmATTACH_POSITION); ac++;
    XtSetArg(al[ac],XmNbottomPosition, 75+y*9); ac++;
    command[y]=XmCreatePushButton(form,"label",al,ac);
    XtManageChild(command[y]);
    XtAddCallback(command[y],XmNactivateCallback,
      commandCB,(XtPointer) cmd_number);
  }

  /* Create a separator widget and attach it to the form */

  ac=0;
  XtSetArg(al[ac],XmNtopAttachment,XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNtopPosition,86); ac++;
  XtSetArg(al[ac],XmNrightAttachment,XmATTACH_FORM); ac++; 
  XtSetArg(al[ac],XmNleftAttachment,XmATTACH_FORM); ac++;  
  XtSetArg(al[ac],XmNbottomAttachment,XmATTACH_NONE); ac++;      
  sep_status=XmCreateSeparator(form,"sep",al,ac);
  XtManageChild(sep_status);   


  /* Message area*/

  /* Create a message widget and attach it to the form */

  ac=0;
  XtSetArg(al[ac],XmNleftAttachment,
    XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNleftPosition,8); ac++;
  XtSetArg(al[ac],XmNrightAttachment,
    XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNrightPosition,92); ac++;
  XtSetArg(al[ac],XmNtopAttachment,
    XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNtopPosition,89); ac++;
  XtSetArg(al[ac],XmNbottomAttachment,
    XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNbottomPosition,98); ac++;
  XtSetArg(al[ac],XmNeditable,False); ac++;
  XtSetArg(al[ac],XmNeditMode,XmMULTI_LINE_EDIT); ac++;
  message_area=XmCreateScrolledText(form,"message_area",al,ac);
  XtManageChild(message_area);

  /* Create a prompt dialog for target input but leave it unmanaged */

  ac=0;
  XtSetArg(al[ac], XmNselectionLabelString,
    XmStringCreateLtoR("Enter coordinates: ",char_set)); ac++;
  XtSetArg(al[ac], XmNdialogTitle,
    XmStringCreateLtoR("XmTel: target",char_set)); ac++;
  dialog=XmCreatePromptDialog(toplevel,"dialog",al,ac);
  XtAddCallback(dialog, XmNokCallback, dialogCB, (int *) OK);
  XtAddCallback(dialog, XmNcancelCallback, dialogCB, (int *) CANCEL);
  XtUnmanageChild (XmSelectionBoxGetChild (dialog, XmDIALOG_HELP_BUTTON));

  /* Create a dialog to select the log file but leave it unmanaged */

  ac=0;
  XtSetArg(al[ac],XmNmustMatch,True); ac++;
  XtSetArg(al[ac],XmNautoUnmanage,False); ac++;
  select_logfile=XmCreateFileSelectionDialog(toplevel,"select_logfile",al,ac);
  XtAddCallback (select_logfile, XmNokCallback, selectlogCB, (int *) OK);
  XtAddCallback (select_logfile, XmNcancelCallback, selectlogCB, (int *) CANCEL);
  XtUnmanageChild(XmSelectionBoxGetChild(select_logfile,
    XmDIALOG_HELP_BUTTON));
  
  /* Create a dialog to select and read the queue but leave it unmanaged */

  ac=0;
  XtSetArg(al[ac],XmNmustMatch,True); ac++;
  XtSetArg(al[ac],XmNautoUnmanage,False); ac++;
  read_queuefile=XmCreateFileSelectionDialog(toplevel,"read_queuefile",al,ac);
  XtAddCallback (read_queuefile, XmNokCallback, readqueueCB, (int *) OK);
  XtAddCallback (read_queuefile, XmNcancelCallback, readqueueCB, (int *) CANCEL);
  XtUnmanageChild(XmSelectionBoxGetChild(read_queuefile,
    XmDIALOG_HELP_BUTTON));
      
  /* Create the menubar */
  
  create_menus(menu_bar);  
  
  /* Connect to XEphem's goto fifo */

  if (fd_fifo_in > 0)
  {
    XtAppAddInput (context, fd_fifo_in, (XtPointer)XtInputReadMask,
      read_xephem_target, NULL);
  }       

  
  /* Create unmanaged control panels */

  setup_model_edit();

  /* Service status display and slewing poll */

  poll_interval_id = XtAppAddTimeOut(context,
    poll_interval,poll_interval_handler,poll_interval_data_ptr);    
  
  /* Start the user interface and telescope communications */
  
  XtRealizeWidget(toplevel);

  XtAppMainLoop(context);

  exit(0);
}

/* Add an item into a menu in the menubar */     

Widget make_menu_item(char *item_name, unsigned long client_data, Widget menu)

{
  int ac;
  Arg al[16];
  Widget item;

  ac = 0;
  XtSetArg(al[ac], XmNlabelString,
      XmStringCreateLtoR(item_name,char_set)); ac++;      
  item=XmCreatePushButton(menu,item_name,al,ac);
  XtManageChild(item);
  XtAddCallback (item,XmNactivateCallback,menuCB,(XtPointer) client_data);      
  XtSetSensitive(item,True);
  return(item);
}

/* Add a toggle item into a menu in the menubar */     

Widget make_menu_toggle(char *item_name, unsigned long client_data, Widget menu)

{
  int ac;
  Arg al[16];
  Widget item;

  ac = 0;
  XtSetArg(al[ac], XmNlabelString,
      XmStringCreateLtoR(item_name,char_set)); ac++;      
  item=XmCreateToggleButton(menu,item_name,al,ac);
  XtManageChild(item);
  XtAddCallback (item,XmNvalueChangedCallback,menuCB,(XtPointer) client_data);      
  XtSetSensitive(item,True);
  return(item);
}

/* Creates a menu on the menu bar */

Widget make_menu(char *menu_name, Widget menu_bar)
 
{
  int ac;
  Arg al[16];
  Widget menu, cascade;

  menu=XmCreatePulldownMenu(menu_bar,menu_name,NULL,0);
  ac=0;
  XtSetArg(al[ac],XmNsubMenuId,menu);  ac++;
  XtSetArg(al[ac],XmNlabelString,
      XmStringCreateLtoR(menu_name,char_set)); ac++;
  cascade=XmCreateCascadeButton(menu_bar,menu_name,al,ac);      
  XtManageChild (cascade); 
  return(menu);
}

/* Create all the menubar menus */

void create_menus(Widget menu_bar)
{
  /* Create the file menu */
  file_menu=make_menu("File",menu_bar);
  newqueue_item=make_menu_item("New Queue",NEWQUEUE,file_menu); 
  newlog_item=make_menu_item("New Log",NEWLOG,file_menu); 
  exit_item=make_menu_item("Exit XmTel",EXIT,file_menu);

  /* Create the edit menu */
  edit_menu=make_menu("Edit",menu_bar);
  editqueue_item=make_menu_item("Queue",EDITQUEUE,edit_menu);
  editlog_item=make_menu_item("Log",EDITLOG,edit_menu);
  
  /* Create the pointing menu with toggle options */
  pointing_menu=make_menu("Pointing",menu_bar);
  p_options_toggle_1  = make_menu_toggle("Offset",POINTOPTION1,pointing_menu);
  p_options_toggle_2  = make_menu_toggle("Refraction",POINTOPTION2,pointing_menu);
  p_options_toggle_4  = make_menu_toggle("Polar",POINTOPTION4,pointing_menu);
  p_options_toggle_8  = make_menu_toggle("Model",POINTOPTION8,pointing_menu);

  /* Create the reference pull-down menu */
  ref_menu            = make_menu("Reference",menu_bar);
  ref_target_item     = make_menu_item("Target",REFTARGET,ref_menu);
  ref_clear_item      = make_menu_item("Clear",REFCLEAR,ref_menu);

  /* Create the model pull-down menu */
  model_menu          = make_menu("Model",menu_bar);
  model_edit_item     = make_menu_item("Edit",MODELEDIT,model_menu);
  model_clear_item    = make_menu_item("Clear",MODELCLEAR,model_menu);
  model_default_item  = make_menu_item("Default",MODELDEFAULT,model_menu);
   
}

/* Callback routines used for menubar */

void menuCB(w,client_data,call_data)
  Widget w;
  int client_data;
  XmAnyCallbackStruct *call_data;
{

  /* If request for a queue file detected, then select a new queue file */

  if (client_data==NEWQUEUE)
  {
    /* Make the queue selection dialog appear */
    
    XtManageChild(read_queuefile);    
  }
  
  /* If request for a new log file detected, then select a new log file */

  if (client_data==NEWLOG)
  {
    /* Make the log selection dialog appear */
    
    XtManageChild(select_logfile);
  } 
  
   
  /* If exit detected, then exit leaving xpatel running */
  
  if (client_data==EXIT) 
  {
    unlink_fifos();
    exit(0);
  }   

  /* If request to edit the queue is detected, then edit the queue file */

  if (client_data==EDITQUEUE)
  {
    strcpy(buf, editor);
    strcat(buf, " ");
    strcat(buf, queuefile);
    system(buf); 
  } 
  
  /* If request to edit log file detected, then edit the log file */

  if (client_data==EDITLOG)
  {
    strcpy(buf, editor);
    strcat(buf, " ");
    strcat(buf, logfile);
    system(buf); 
  }
  
  /* Detect a change to the toggle for a pointing option */
  /* Check for the state of the option                   */
  /* Set pmodel using bitwise OR or bitwise AND NOT      */
  /* Set the appropriate telxpa state */

  if (client_data==POINTOPTION1)
  {
    if( XmToggleButtonGetState(p_options_toggle_1) )
    {
      pmodel=(pmodel | OFFSET);
    }
    else
    {
      pmodel=(pmodel & ~(OFFSET));
    }      
    tel_send_offset();
  } 
      
  if (client_data==POINTOPTION2)
  {
    if( XmToggleButtonGetState(p_options_toggle_2) )
    {
      pmodel=(pmodel | REFRACT);
    }
    else
    {
      pmodel=(pmodel & ~(REFRACT));      
    }
    tel_send_refract();
  }         

  if (client_data==POINTOPTION4)
  {
    if( XmToggleButtonGetState(p_options_toggle_4) )
    {
      pmodel=(pmodel | POLAR);
    }
    else
    {
      pmodel=(pmodel & ~(POLAR));
    }  
    tel_send_polar();
  }         

  if (client_data==POINTOPTION8)
  {
    if( XmToggleButtonGetState(p_options_toggle_8) )
    {
      pmodel=(pmodel | DYNAMIC);
    }
    else
    {
      pmodel=(pmodel & ~(DYNAMIC));
    }
    tel_send_dynamic();  
  }         

  /* Reference to target request detected */

  if (client_data==REFTARGET)
  {
    tel_send_offset_target();
    tel_receive_telescope();
    show_telescope_coordinates();
    mark_xephem_telescope();
  }       

  /* Reference to clear request detected */

  if (client_data==REFCLEAR)
  {
    tel_send_offset_clear();
    tel_receive_telescope();
    show_telescope_coordinates();
    mark_xephem_telescope();    
  }       

  /* If model edit is requested, then handle that here */
  
  if (client_data==MODELEDIT)
  {
    /* Make the model edit panel appear */
    
    XtManageChild(model_edit);    
  }         

  /* If model clear detected, then update offset */

  if (client_data==MODELCLEAR)
  {
     modelha1 = 0.;
     modeldec1 = 0.;
     
  } 
    
  /* If model default detected, then update offset */
  /* Default is set to 0. on startup then reset by the configuration file */

  if (client_data==MODELDEFAULT)
  {

  } 

  /* Else noop */   
   
  if (client_data==CANCEL);
}

/* Create the unmanaged mount model edit panel */

void setup_model_edit()
{
  Arg al[10];
  int ac;
  
  ac=0;
  
  XtSetArg(al[ac], XmNselectionLabelString,
    XmStringCreateLtoR("Enter mount model: ",char_set)); ac++;
  XtSetArg(al[ac], XmNdialogTitle,
    XmStringCreateLtoR("Edit mount model",char_set)); ac++;
  model_edit = XmCreatePromptDialog(toplevel,"model_edit",al,ac);
  XtAddCallback(model_edit, XmNokCallback, model_editCB, (XtPointer) OK);
  XtAddCallback(model_edit, XmNcancelCallback, model_editCB, (XtPointer) CANCEL);
  XtUnmanageChild (XmSelectionBoxGetChild (model_edit, XmDIALOG_HELP_BUTTON));
}

/* Callback function for the mount model edit box */
/* Read mount model in units of pixels/hour       */
/* Units conversion done in pointing utility */
     
void model_editCB(w,client_data,call_data)
  Widget w;
  int client_data;
  XmSelectionBoxCallbackStruct  *call_data;
{    
  char *s;
  int testflag;
  double tmpmodelha1, tmpmodeldec1;
  
  switch (client_data)
  {
    case OK:
      XmStringGetLtoR(call_data->value,char_set,&s);
      testflag = sscanf(s, "%lf %lf ", &tmpmodelha1, &tmpmodeldec1);
      if (testflag > 0)
      {
          modelha1 = tmpmodelha1;
          modeldec1 = tmpmodeldec1;            
          tel_send_dynamic();
      }
            
      XtFree(s);
      break;
    case CANCEL:
       break;
  }
  XtUnmanageChild(w);
}

      
/* Callback function for the target input box */

void dialogCB(w,client_data,call_data)
  Widget w;
  int client_data;
  XmSelectionBoxCallbackStruct  *call_data;
{    
  char *s;
  double tmpra;
  double tmpdec;
  int testflag;

  switch (client_data)
  {
    case OK:
      XmStringGetLtoR(call_data->value,char_set,&s);
      switch (target)
      {  
        case RA: 
        testflag = dmstod(s,&tmpra);
	if (testflag == 0)
	{
	  if (display_targetepoch == EOD)
	  {
	    /* Epoch of date coordinates apply to entry */
            targetra = tmpra;
	    tmpdec = targetdec;
	    Apparent(&tmpra, &tmpdec, -1);
	    targetra2 = tmpra;
	    targetdec2 = tmpdec;	    
	  }
	  else
	  {
	    /* Epoch 2000.0 coordinates apply to entry */
	    targetra2 = tmpra;
	    tmpdec = targetdec2;
	    Apparent(&tmpra, &tmpdec, 1);
	    targetra = tmpra;
	    targetdec = tmpdec;
	  }
	}
        break;
        
        case DEC: 
        testflag = dmstod(s,&tmpdec);
	if (testflag == 0)
	{	
	  if (display_targetepoch == EOD)
	  {
	    /* Epoch of date coordinates apply to entry */  
	    targetdec = tmpdec; 
	    tmpra = targetra;
	    Apparent(&tmpra, &tmpdec, -1);	    	  
	    targetra2 = tmpra;
	    targetdec2 = tmpdec;	    
	  }
	  else
	  {
	    /* Epoch 2000.0 coordinates apply to entry */ 
	    targetdec2 = tmpdec;   
	    tmpra = targetra2;
	    Apparent(&tmpra, &tmpdec, 1);
	    targetra = tmpra;
	    targetdec = tmpdec;	         
	  }
	}
        break;
        
      }
      XtFree(s);
      break;
    case CANCEL:
       break;
  }
  tel_send_target();
  tel_receive_target();
  show_target_coordinates();
  mark_xephem_target(); 
    
  XtUnmanageChild(w);
}

/* Callback function for the speed toggle buttons */

void speedCB(w,client_data,call_data)
  Widget w;
  int client_data;
  XmAnyCallbackStruct *call_data;
{
  int n;

  /* Identify the button code*/
  n=client_data;
     
  /* Assign a speed code */
  
   switch (n) 
   {
  
     /* Slew */
     case 0: 
       telspd=SLEW;
       tel_send_speed();
       break; 
        
     /* Find */
     case 1: 
       telspd=FIND;
       tel_send_speed();
       break;
         
     /* Center */
     case 2: 
       telspd=CENTER;
       tel_send_speed();
       break; 
      
     /* Guide */
     case 3: 
       telspd=GUIDE;
       tel_send_speed();
       break;  
   } 
}

/* Callback function for the 9 grid buttons. */
/* Called when one of the grid buttons is pressed. */

void buttonCB(w,client_data,call_event)
  Widget w;
  int client_data;
  XEvent *call_event;
{
  int n;
  int et = call_event->type;
  int dn,up;
   
  /* identify the button number n = 3*x+y */
  n=client_data;

  /* identify the action */
  
  dn = et == ButtonPress;
  up = et == ButtonRelease;
      
  /* respond to the request */

  if(dn)
  {

    switch (n) 
    {
      /* NoOp */
      case 0:
      break;

      /* E */
      case 1: 
        teldir = EAST;  
	movetime = 0.5*((double) telspd);
        tel_send_move();
        strcpy(message,"Moving east\n");
        show_message();
      break;

      /* NoOp */
      case 2: 
      break;

      /* N */
      case 3: 
        teldir = NORTH;
	movetime = 0.5*((double) telspd);
        tel_send_move();
        strcpy(message,"Moving north\n");
        show_message();
      break;

      /* NoOp */
      case 4:
      break;

      /* S */
      case 5: 
        teldir = SOUTH;
	movetime = 0.5*((double) telspd);
        tel_send_move();
        strcpy(message,"Moving south\n");
        show_message();
      break;

      /* NoOp */
      case 6:
      break;

      /* W */
      case 7:
        teldir = WEST;
	movetime = 0.5*((double) telspd);
        tel_send_move();
        strcpy(message,"Moving west\n");
        show_message();
      break;

      /* NoOp */
      case 8: 
      break;    
   

    }
  }
  
  if(up)
  {

    switch (n) 
    {
      /* NoOp */
      case 0:
      break;

      /* E */
      case 1: 
        teldir = EAST;  
	movetime = 0.;
        tel_send_move();
        strcpy(message,"Stopping east\n");
        show_message();
      break;

      /* NoOp */
      case 2: 
      break;

      /* N */
      case 3: 
        teldir = NORTH;
	movetime = 0.;
        tel_send_move();
        strcpy(message,"Stopping north\n");
        show_message();
      break;

      /* NoOp */
      case 4:
      break;

      /* S */
      case 5: 
        teldir = SOUTH;
	movetime = 0.;
        tel_send_move();
        strcpy(message,"Stopping south\n");
        show_message();
      break;

      /* NoOp */
      case 6:
      break;

      /* W */
      case 7:
        teldir = WEST;
	movetime = 0.;
        tel_send_move();
        strcpy(message,"Stopping west\n");
        show_message();
      break;

      /* NoOp */
      case 8: 
      break;    
   
    }


    /* Update the coordinate displays */
                 
    tel_receive_telescope();
    show_telescope_coordinates();
    mark_xephem_telescope();


    /* Update the console message */

    if (guideflag == TRUE)
    {
      strcpy(message,"Guiding\n");
      show_message();    
    }
    else if (trackflag == TRUE)
    {
      trackflag = FALSE;
      gotoflag = FALSE;
      tel_send_track();
      tel_send_goto();
      trackflag = TRUE;
      tel_send_track(); 
      strcpy(message,"Tracking\n");
      show_message();
    }  
    else if (gotoflag == TRUE)
    {
      strcpy(message,"Slewing\n");
      show_message();
    }  
    else
    {
      trackflag = FALSE;
      gotoflag = FALSE;
      tel_send_track();
      tel_send_goto();
      strcpy(message,"Paused\n");
      show_message();
    }  
  }
}


/* Callback function for the 8 telescope data command buttons */
/* All of them respond with a mark on XEphem */
/* Epoch switch toggles the epoch of the displayed coordinates */

void teldataCB(w,client_data,call_data)
  Widget w;
  int client_data;
  XmAnyCallbackStruct *call_data;
{
   
  int n, cmd_number;

  /* Identify the button */
  cmd_number=client_data;

  /* Respond to the request */
  n=cmd_number;

  switch (n) 
  {

    /* Read telescope coordinates from xpatel, show and mark on XEphem */
      case 0: 
      tel_receive_telescope();
      show_telescope_coordinates();
      mark_xephem_telescope();
      break;

    /* Read target coordinates from xpatel, show and mark on XEphem */
    case 1: 
      tel_receive_target();
      show_target_coordinates();
      mark_xephem_target();
      break;

    /* Read telescope coordinates from xpatel, show and mark on XEphem */
    case 2:
      show_telescope_coordinates();
      mark_xephem_telescope();
      break;

    /* Read telescope coordinates from xpatel, show and mark on XEphem */
    case 3: 
      show_telescope_coordinates();
      mark_xephem_telescope();
      break;
      
    /* Update telescope display epoch */
    case 4: 
      display_telepoch = display_telepoch + 1;
      display_telepoch = display_telepoch % 2;
      show_telescope_coordinates();
      mark_xephem_telescope();
      break;        
      
    /* Enter target RA from the console */
    case 5: 
      new_target_ra();
      tel_receive_target();
      show_target_coordinates();
      mark_xephem_target(); 
      break;

    /* Enter target Dec from the console */
    case 6: 
      new_target_dec();
      tel_receive_target();
      show_target_coordinates();
      mark_xephem_target();
      break;

    /* Update target display epoch */
    case 7: 
      display_targetepoch = display_targetepoch + 1;
      display_targetepoch = display_targetepoch % 2;
      show_target_coordinates();
      break;
  }
}

/* Callback function for the 8 telescope command buttons */

void commandCB(w,client_data,call_data)
  Widget w;
  int client_data;
  XmAnyCallbackStruct *call_data;
{
  int n, cmd_number;

  /* identify the button */
  cmd_number=client_data;

  /* respond to the request */
  n=cmd_number;

  switch (n) 
  {

    /* Slew to target */
    case 0: 
      /* If guiding turn this feature off first */
      
      if(guideflag == TRUE)
      {
        guideflag = FALSE;
        tel_send_guide();
      }    
      gotoflag = TRUE;
      tel_send_goto();
      strcpy(message,"Slewing\n");
      show_message();
      
      break;     

    /* Save telescope coordinates to the log */
    case 1: 
      save_coordinates();
      break;

    /* Tracking on */
    case 2: 
      gotoflag = FALSE;
      trackflag = TRUE;
      guideflag = FALSE;
      tel_send_goto();
      tel_send_track();
      tel_send_guide();
      strcpy(message,"Tracking on\n");
      show_message();
      break; 
 
    /* Recall a target from the log */
    case 3: 
      recall_coordinates();
      break;

    /* Precision closed loop guide */
    case 4: 
      if(trackflag == FALSE)
      {
        trackflag = TRUE;
        tel_send_track();
      }
      guideflag = TRUE;
      tel_send_guide();
      strcpy(message,"Tracking on and guiding on\n");
      show_message();
      break;        

    /* Load home position as target */
    case 5:       
      /* If guiding turn this feature off first */
      
      if(guideflag == TRUE)
      {
        guideflag = FALSE;
        tel_send_guide();
      }    
      homeflag = TRUE;
      parkflag = FALSE;
      tel_send_home();
      tel_receive_target();
      show_target_coordinates();
      mark_xephem_target();
      break;     



    /* Stop all motion */
    case 6: 
      trackflag = FALSE;
      gotoflag = FALSE;
      guideflag = FALSE;
      homeflag = FALSE;
      parkflag = FALSE;
      tel_send_track();
      tel_send_guide();
      tel_send_goto();
      strcpy(message,"Tracking and guiding off\n");      
      show_message();
      break; 

    /* Load park position as the  target*/
    case 7: 
      /* If guiding turn this feature off first */
      
      if(guideflag == TRUE)
      {
        guideflag = FALSE;
        tel_send_guide();
      }    
      parkflag = TRUE;
      homeflag = FALSE;
      tel_send_park();
      tel_receive_target();
      show_target_coordinates();
      mark_xephem_target();

      break;     
 
  }
}

/* Callback function for selecting the log file name */

void selectlogCB(w,client_data,call_data)
  Widget w;
  int client_data;
  XmAnyCallbackStruct *call_data;
 
{
  XmFileSelectionBoxCallbackStruct *s =
      (XmFileSelectionBoxCallbackStruct *) call_data;
  
  /* Do nothing if cancel is selected. */
  if (client_data==CANCEL) 
  {
      XtUnmanageChild(select_logfile);
      return;
  }

  /* Manage memory for file name */
  
  if(logfile != NULL)
  {
    XtFree(logfile);
    logfile = NULL;
  }
  
  /* Get the filename from the file selection box */
  XmStringGetLtoR(s->value, char_set, &logfile);
  XtUnmanageChild(select_logfile);
} 
 


/* Callback function for selecting and reading the queue file  */
 
void readqueueCB(w,client_data,call_data)
  Widget w;
  int client_data;
  XmAnyCallbackStruct *call_data;
{
  XmFileSelectionBoxCallbackStruct *s =
      (XmFileSelectionBoxCallbackStruct *) call_data;
  
  /* Do nothing if cancel is selected. */
  if (client_data==CANCEL) 
  {
    XtUnmanageChild(read_queuefile);
    return;
  }

  /* Manage memory for file name */
  if(queuefile != NULL)
  {
    XtFree(queuefile);
    queuefile = NULL;
  }
  
  /* Get the filename from the file selection box */
  XmStringGetLtoR(s->value, char_set, &queuefile);

  /* Read the queue into memory */
  read_queue();

  XtUnmanageChild(read_queuefile);

}    

/* Callback function for selecting a queue entry */

void selectqueueCB(w,client_data,call_data)
  Widget w;
  XtPointer client_data;
  XtPointer call_data;
{
  XmListCallbackStruct *cbs = (XmListCallbackStruct *) call_data;

  /* Selection is indexed from 1 in XmList and from 0 in our list */
  
  queuechoice =  cbs->item_position - 1;    
    
  if (!quiet)
  {
    printf(" Queue entry #%d: \n",queuechoice);                                          
    printf("   %s\n",queue[queuechoice].name);    
    printf("   ra: %lf  dec: %lf\n\n",                             
    queue[queuechoice].ra,queue[queuechoice].dec);  
  }

  targetra2 = queue[queuechoice].ra;
  targetdec2 = queue[queuechoice].dec;
  targetra = targetra2;
  targetdec = targetdec2;
  Apparent(&targetra, &targetdec, 1);
  tel_send_target();
  tel_receive_target();
  show_target_coordinates();
  mark_xephem_target(); 
}    


/* Show latest message on the user interface */

void show_message()
{
  XmString xmstr;
  
  xmstr = XmStringCreateLtoR(message,char_set);
  XmTextSetString(message_area, message); 
  XmStringFree(xmstr);
} 

/* Show telescope coordinates for telepoch in the button */

void show_telescope_coordinates()
{
  double tmpra, tmpdec;
  XmString xmstr;
  Arg al[10];
  int ac;

  switch (display_telepoch)
  {
          
    case J2000:
    
      /* Display Epoch 2000.0 coordinates */      
    
      xmstr = XmStringCreateLtoR("2000.0",char_set);
      ac=0; 
      XtSetArg(al[ac],XmNlabelString,xmstr); ac++; 
      XtSetValues(get_telepoch,al,ac);            
      tmpra = telra2;
      tmpdec = teldec2;
      
      break;
    
    case  EOD:
    
      /* Display EOD coordinates */      
    
      xmstr = XmStringCreateLtoR("Now",char_set);
      ac=0; 
      XtSetArg(al[ac],XmNlabelString,xmstr); ac++; 
      XtSetValues(get_telepoch,al,ac);
      tmpra=telra;
      tmpdec=teldec; 
      break;
  }   

  dtodms(buf, &tmpra);
  xmstr = XmStringCreateLtoR(buf,char_set);
  ac=0;
  XtSetArg(al[ac],XmNlabelString,xmstr); ac++;
  XtSetValues(get_telra,al,ac);
  
  dtodms(buf, &tmpdec);
  xmstr = XmStringCreateLtoR(buf,char_set);
  ac=0;
  XtSetArg(al[ac],XmNlabelString,xmstr); ac++;
  XtSetValues(get_teldec,al,ac);    

  XmStringFree(xmstr);
  
}  
  
/* Show target coordinates for targetepoch in the button */

void show_target_coordinates()
{
  double tmpra, tmpdec;
  XmString xmstr;
  Arg al[10];
  int ac;

  switch (display_targetepoch)
  {
    
    case J2000:
    
      /* Display epoch 2000.0 target coordinates */
    
      ac=0;      
      xmstr = XmStringCreateLtoR("2000.0",char_set);
      XtSetArg(al[ac],XmNlabelString,xmstr); ac++; 
      XtSetValues(get_targetepoch,al,ac);
      tmpra=targetra2;
      tmpdec=targetdec2;        
      break;
    
    case  EOD:
    
      /* Display EOD target coordinates */
    
      xmstr = XmStringCreateLtoR("Now",char_set);
      ac=0; 
      XtSetArg(al[ac],XmNlabelString,xmstr); ac++; 
      XtSetValues(get_targetepoch,al,ac);
      tmpra=targetra;
      tmpdec=targetdec;          
      break;
  }      
 
  dtodms(buf, &tmpra);
  xmstr = XmStringCreateLtoR(buf,char_set);
  ac=0;
  XtSetArg(al[ac],XmNlabelString,xmstr); ac++;
  XtSetValues(get_targetra,al,ac);   

  dtodms(buf, &tmpdec);
  xmstr = XmStringCreateLtoR(buf,char_set);
  ac=0;
  XtSetArg(al[ac],XmNlabelString,xmstr); ac++; 
  XtSetValues(get_targetdec,al,ac);       

  XmStringFree(xmstr);
}  

void poll_interval_handler(XtPointer client_data_ptr, XtIntervalId *client_id) 
{
  
  /* If not initialized say Hello */
  
  if ( telinitflag == FALSE )
  {
    telinitflag = TRUE;
    
    /* Tell the operator that the telescope is ready */
     
    strcpy(message,"Telescope is ready\n");

    /* Update the coordinate displays */
  	       
    tel_receive_telescope();
    show_telescope_coordinates();
    mark_xephem_telescope();	 
  } 

  /* Manage user inferface updates promptly */ 
  
  show_message();

            
  /* Check the slew status */

  if ( gotoflag == TRUE )
  {
     tel_receive_goto();
     if ( gotoflag == FALSE )
     {
       /* Restore the user-selected speed in case it was changed */
       
       tel_send_speed();
       
       /* Tell the operator that the slew has been completed */
       
      strcpy(message,"Slew to target completed\n");
      show_message();

      /* Update the coordinate displays */
                 
      tel_receive_telescope();
      show_telescope_coordinates();
      mark_xephem_telescope();

     }

  } 
  
  
  /* Start the timer again.  It will return to the handler after poll_interval. */

  poll_interval_id = XtAppAddTimeOut(context, poll_interval,
    poll_interval_handler, poll_interval_data_ptr);
}  


/* Export current telescope coordinates to XEphem */

void mark_xephem_telescope()     
{
  double tmpra, tmpdec;
  double rarad,decrad;
  char outbuf[101];
  
  /* Recall the telescope's 2000.0 coordinates */
  
  tmpra=telra2;
  tmpdec=teldec2;         

  /* Convert to radians */
  
  rarad=tmpra*PI/12.;  
  decrad=tmpdec*PI/180.;  
  
  /* Send the values to the fifo */
  
  snprintf(outbuf,sizeof(outbuf),"RA:%9.6f Dec:%9.6f Epoch:2000.000\n",
    rarad,decrad);  
  write(fd_fifo_out, outbuf, strlen(outbuf));
}

/* Export current target coordinates to XEphem */

void mark_xephem_target()     
{
  double tmpra, tmpdec;
  double rarad,decrad;
  char outbuf[101]; 
  
  /* Recall the telescope's 2000.0 coordinates */
  
  tmpra=targetra2;
  tmpdec=targetdec2;         
  
  /* Convert to radians */
  
  rarad=tmpra*PI/12.;  
  decrad=tmpdec*PI/180.; 
  
  /* Send the values to the fifo */  
  
  snprintf(outbuf,sizeof(outbuf),"RA:%9.6f Dec:%9.6f Epoch:2000.000\n",
    rarad,decrad);  
  write(fd_fifo_out, outbuf, strlen(outbuf));
}

/* Read XEphem goto fifo and parse target coordinates */

void read_xephem_target() 
{
  int i,j,k;
  double rahr,ramin,rasec;
  double decdeg,decmin,decsec;
  double tmpra,tmpdec,tmpepoch;
  double tmp_pm_ra = 0.;
  double tmp_pm_dec = 0.;
  double tmpdat;
  char fifostr[121];
  char rastr[121];
  char *raptr=rastr;  
  char inbuf[2];
  
  /* A typical fixed target string is:                            */

  /* Regulus,f|M|B7,10:08:22.3, 11:58:02,  1.35,2000,0            */
  /* Name, type info, RA | pm, Dec | pm, Mag, Epoch, Identifier   */

  /* Return if the buffer is not open                             */

  if(fd_fifo_in<0)
    return;

  /* Read the buffer */

  k=0;
  while (1) 
  {
    i = read(fd_fifo_in,inbuf,1);
    if(i==-1) 
    {
      fifostr[k]='\0';
      break;
    }
    if(k<120)
    fifostr[k++]=inbuf[0];
  }
  
  /* Diagnostic of fifo string contents */
  
  /*  printf("Fifo in: %s\n",fifostr); */
    
  /* Examples from XEphem version 3.6.xx -- */
  
  /* Arcturus,f|V|K1,14:15:39.7|-1093, 19:10:57|-1998,-0.04,2000,0  */
  /* HD 124953,f|V|A8,14:16:04.2|43, 18:54:43|-28,5.98,2000,0       */
  /* Crt Delta-12,f|S|G8,11:19:20.5|-122,-14:46:43|208,3.56,2000,0  */
  /* GSC 0838-0788,f|S,10:32:32.4,  9:06:17,13.84,2000,0            */
  /* M67,f|O|T2, 8:51:24.0, 11:49:00,6.90,2000,1500|1500|0          */
  /* NGC 2539,f|O|T2, 8:10:36.9,-12:49:14,6.50,2000,900|900|0       */
  /* Venus,P                                                        */


  /* If the buffer was not empty then extract its contents */

  if (k>1) 
  {
    fifostr[k]='\0';

    /* Remove any white space */

    i=0;
    j=0;
    while (i<=k)
    {
      if( fifostr[i]!=' ' )
      {
        rastr[j]=fifostr[i];
        j++;
      }
      i++;
    }  
    
    raptr = rastr;
    i=0;
    while (i<2)  
    {
      raptr=strchr(raptr,',');
      if(raptr==NULL) break;
      raptr++;
      i++;
    }
        
    /* Parsing will have proceeded to the RA field if this is a catalog object */

    if (i==2)
    { 

      /* Scan and save RA data */ 
      /* Note that the XEphem proper motion is in milliarcseconds per year */ 
      /* Our pm_ra is in seconds of time per year */

      tmpdat = 0.;
      j=sscanf(raptr,"%lf:%lf:%lf|%lf", 
        &rahr,&ramin,&rasec,&tmpdat);

      if(j>0) 
      {
        tmpra = rahr + ramin/60. + rasec/3600.;
        tmp_pm_ra=tmpdat/15000.;
      }
       
      /* Try to advance pointer to the Dec field */

      raptr=strchr(raptr,',');
      if(raptr != NULL)
      {
       raptr++;
       i++;
      }
    }
     

    if (i==3)
    {  

      /* Now pointing at the Dec field so scan and save Dec data */
      /* Note that the XEphem proper motion is in milliarcseconds per year */ 
      /* Our pm_dec is in seconds of arc per year */

      tmpdat = 0.;
      j=sscanf(raptr,"%lf:%lf:%lf|%lf", 
        &decdeg,&decmin,&decsec,&tmpdat);

      if (j>0) 
      {
        if(decdeg>0.)
        {
          tmpdec = decdeg + decmin/60. + decsec/3600.;
        }
        else
        { 
          tmpdec = decdeg - decmin/60. - decsec/3600.;
        }
        tmp_pm_dec=tmpdat/1000.;
      }     

      /* Try to advance pointer to the Magnitude field */

      raptr=strchr(raptr,',');
      if(raptr != NULL)
      {
       raptr++;
       i++;
      }
            
    }

    if (i==4)
    {  

      /* Now pointing at the Magnitude field so scan past it */ 

      j=sscanf(raptr,"%lf", &tmpdat);
      
      /* Magnitude is not used later but if(j>0) magnitude is tmpdat */
    
      /* Try to advance pointer to the Epoch field */

      raptr=strchr(raptr,',');
      if(raptr != NULL)
      {
       raptr++;
       i++;
      }   
    }
  
    if (i==5)
    {  

      /* Now pointing at the Epoch field */ 

      j=sscanf(raptr,"%lf", &tmpepoch);

      if(j>0) 
      {

         /* The catalog entry epoch has been found */
         /* Apply proper motion to EOD */

         ProperMotion(tmpepoch, &tmpra, &tmpdec, tmp_pm_ra, tmp_pm_dec);        
      
         /* Find the apparent coordinates including */
         /*   precession, nutation, and stellar aberration for the EOD */

         if ((tmpepoch > 1999.99) && (tmpepoch < 2000.01)) 
         {
           
           /* Epoch 2000.0 coordinates apply */
           
           targetra2 = tmpra;
           targetdec2 = tmpdec;
           targetra = tmpra;
           targetdec = tmpdec;
           Apparent(&targetra, &targetdec, 1); 
 

         }
         else 
         {
 
           /* Some other epoch applies.  Caution!!  */
           /* Set the epoch to 2000.0 and then proceed as before. */
           /* If coordinates are not FK5 this will not be accurate. */
                     
           PrecessToEOD(tmpepoch, &tmpra, &tmpdec);
           PrecessToEpoch(2000.0, &tmpra, &tmpdec);
          
           /* Epoch 2000.0 coordinates now apply */
           /* Find the apparent position of the object for EOD */

           targetra2 = tmpra;
           targetdec2 = tmpdec;
           targetra = tmpra;
           targetdec = tmpdec;
           Apparent(&targetra, &targetdec, 1); 
           
         }
                           
         /* Send the selection to xpatel */
 
         tel_send_target();        
         
         /* Show the entry at the epoch selected for display */         
         
         show_target_coordinates();
      }
      else
      {
         
         /* No epoch was found so assume E2000 coordinates */
         
         targetra2 = tmpra;
         targetdec2 = tmpdec;     
         
         /* Send the selection to xpatel */
         
         tel_send_target();
                     
         /* Show the entry at the epoch selected for display */
         
         show_target_coordinates();  
      }     
    }  
  }
}

/* Write current telescope coordinates to the xmtel log file and to history */

void save_coordinates()           
{
  char decstr[20], rastr[20];
  time_t timebuf;
  char *gmtstr;
  
  time(&timebuf);
  gmtstr = asctime(gmtime(&timebuf));
  if (gmtstr[strlen(gmtstr) - 1] == '\n') gmtstr[strlen(gmtstr) - 1]= '\0';

  dtodms(rastr,&telra);
  dtodms(decstr,&teldec);
  fp_log=fopen(logfile, "a");
  fprintf(fp_log,"%s  %s  %s\n",gmtstr,rastr,decstr);
  fflush(fp_log);
  fclose(fp_log);
  
  /* Expand the buffer up to a maximum of 100 entries */
  
  if (nhistory < 100) 
  {
    nhistory++;
  }
  
  /* Increment the ring counter */
  
  last_history_saved++;
  
  /* Test for wraparound and reset history counters */
  
  if (last_history_saved > 99) 
  {
    last_history_saved=0;
  }
  last_history_recalled = last_history_saved - 1;
  
  /* Save at the current telescope coordinates in the last_history location */
  
  strncpy(history[last_history_saved].name,gmtstr,79);
  history[last_history_saved].ra = telra;
  history[last_history_saved].dec = teldec;

}

/* Read a new target from the xmtel history file */

void recall_coordinates()           
{
  
  if ( last_history_saved >= 0 )
  {
    last_history_recalled++;
    if (last_history_recalled > nhistory -1 )
    {
      last_history_recalled = 0;
    }
    targetra = history[last_history_recalled].ra;
    targetdec = history[last_history_recalled].dec;
    targetra2 = targetra;
    targetdec2 = targetdec;
    Apparent(&targetra2, &targetdec2, -1);
    tel_send_target();
    show_target_coordinates();
    strcpy(message,history[last_history_recalled].name);
    strcat(message,"\n");
    show_message();
  }
  else
  {
    strcpy(message,"Nothing yet saved.\n");
    show_message();
  }
}


/* Read the queue line by line */

void read_queue()
{
  char queuestr[121];
  char *queueptr=queuestr;
  int i,j,k;
  int nline;
  int success;
  double rahr,ramin,rasec;
  double decdeg,decmin,decsec;

  Arg al[10];
  int ac;
  
  strcpy(message,"Reading the queue ");
  strcat(message,queuefile);
  strcat(message,"\n");
  show_message();
  
  fp_queue=fopen(queuefile, "r");

  /* Queue is indexed from 0 so we start with -1 here */
  /* Line counter is indexed from 1 and we start with 0 here */
  
  nqueue = -1;
  nline = 0;
  while ( queuestr == fgets(queuestr,120,fp_queue) )
  {
    /* Reset a flag to determine whether to use the entry */
    
    success = FALSE;
    
    if (!quiet)
    {
    
      printf("Queue line %d: %s\n", nline, queuestr);

    }
     
    /* If the buffer is not empty and not a comment then extract its parts */

    k=strlen(queuestr);
    if ( (k>1) && ( queuestr[0] != '#' ) ) 
    {

      success = TRUE;
      nqueue++;
      nline++;
                    
      /* Null terminate the string */
      
      queuestr[k]='\0';
      queueptr=queuestr;
      
      /* The name field may have useful white space */
      
      /* Identify the entire name field and copy it to the local queue */

      j=strcspn(queueptr,",");
      strncpy(queue[nqueue].name,queueptr,j);
      queueptr = queueptr + j + 1;
      
      /* Remove any white space as the entire string is copied on itself */

      i=0;
      j=0;
      while (i<=k)
      {
        if( queuestr[i]!=' ' )
        {
          queuestr[j]=queuestr[i];
          j++;
        }
        i++;
      }  

      /* Set the pointer at the beginning of the clean string */

      queueptr=queuestr;
      i = 0;
      
      /* Skip the cleaned name field */

      queueptr=strchr(queueptr,',');
      if ((queueptr == NULL) || (queueptr == queuestr))
      {
        success = FALSE;
        i = 0;
      }
      else
      {
        i = 1;
        queueptr++;
      }
      
      if (!quiet)
      {    
        printf("Queue coordinate success flag: %d\n", success);
      }


      /* Parsing may have proceeded to the RA field if this is an object */

      if ( (i == 1) && (success == TRUE) )
      { 

        /* Scan and save RA data in a temporary buffer */ 

        j=sscanf(queueptr,"%lf:%lf:%lf", 
          &rahr,&ramin,&rasec);

        if (j>0) 
        {
          queue[nqueue].ra = rahr + ramin/60. + rasec/3600.;
          
          /* Try to advance pointer to the Dec field */

          queueptr=strchr(queueptr,',');
          if(queueptr != NULL)
          {
            queueptr++;
            i++;
          }
        }
        else
        {
          success = FALSE;
        }                  
      }
     
     
      if ( (i == 2) && (success == TRUE) )
      {  

        /* Now pointing at the Dec field so scan and save Dec data */
        /* Queue proper motion is in milliarcseconds per year */ 
        /* Our pm_dec is in seconds of arc per year */

        j=sscanf(queueptr,"%lf:%lf:%lf", 
        &decdeg,&decmin,&decsec);

        if (j>0) 
        {
          if (decdeg>0.)
          {
            queue[nqueue].dec = decdeg + decmin/60. + decsec/3600.;
          }
          else
          { 
            queue[nqueue].dec = decdeg - decmin/60. - decsec/3600.;
          }
        }
        else
        {        
          success = FALSE;
        }     
      }
               
      if (!success)
      {
        if (!quiet )
        {
          printf(" Skipping queue line %d:\n", nline);                                   
        }
        nqueue--;
      }  
      else 
      {
                                                                    
        if (!quiet )
        {
          printf(" #%d\n",nqueue);                                   
          printf(" %s\n",queue[nqueue].name);  
          printf(" ra: %lf  dec: %lf \n\n",                 
            queue[nqueue].ra, queue[nqueue].dec);   
        }
      } 
    }

  } 

  fclose(fp_queue);

  /* Increment nqueue so that it now reads the number of entries to the queue */
  
  nqueue++;

  XmStringTable queue_list =
        (XmStringTable) XtMalloc ( nqueue * sizeof (XmString) ); 
  for (i=0; i < nqueue; i++)
  {
    queue_list[i] = XmStringCreateLocalized (queue[i].name); 
  }
  ac=0; 
  XtSetArg(al[ac],XmNitemCount,nqueue); ac++; 
  XtSetArg(al[ac],XmNitems,queue_list); ac++;
  XtSetValues(queue_area,al,ac);  
  for (i = 0; i < nqueue; i++)
  {
    XmStringFree(queue_list[i]);
  }
  free(queue_list);

}

/* User input of target RA */

void new_target_ra() 
{
  target=RA;
  XtManageChild(dialog);
}

/* User input of target Dec */

void new_target_dec()
{
  target=DEC;
  XtManageChild(dialog);
}

/* Show new model parameters in the message window */

void show_model_parameters(void)
{
  sprintf(message,"Model parameters for HA: %8.3f  Dec: %8.3f\n", modelha1, modeldec1);      
  show_message();
}

     
/* Start fifo links */

void link_fifos()
{  
  fd_fifo_in = open("/usr/local/xephem/fifos/xephem_loc_fifo", 
    O_RDWR | O_NONBLOCK);


  if (fd_fifo_in<0) {
    printf("Unable to open /usr/local/xephem/fifos/xephem_loc_fifo. \n");
    }
  fd_fifo_out = open("/usr/local/xephem/fifos/xephem_in_fifo", 
    O_RDWR | O_NONBLOCK);


  if (fd_fifo_out<0) {
    printf("Unable to open /usr/local/xephem/fifos/xephem_in_fifo. \n");
    }
}

/* Stop fifo links */

void unlink_fifos()
{
  if (fd_fifo_in!=-1)
    close(fd_fifo_in);
  if (fd_fifo_out!=-1)
    close(fd_fifo_out);  
}

/* Convert string deg:min:sec or hr:min:sec to a double */

int dmstod (char *instr, double *datap)
{
  double h=0., m=0., s=0.;
  int negflag;
  int convertflag;
  
  while (isspace(*instr))
    instr++;
  if (*instr == '-') {
    negflag = 1;
    instr++;
  } else
    negflag = 0;
    
  convertflag = sscanf (instr,  "%lf%*[:]%lf%*[:]%lf", &h, &m, &s);
  if (convertflag < 1)
    return (-1);
  *datap = h + m/60. + s/3600.;
  if (negflag)
    *datap = - *datap;
  return (0);
}

    
/* Convert double to string deg:min:sec or hr:min:sec */
/* This routine has memory leaks that should be fixed */

void dtodms (char *outstr, double *dmsp)
{

  int d=0, m=0, s=0;
  int negflag;
  double dms, ms;
   
  dms = *dmsp;

  /* If negative, treat as positive and set sign in string */

  if (dms < 0 )
  {
    negflag = 1;
    dms = - dms;
  }
  else
  {
    negflag = 0;
  }

  /* Allow for truncation if input is in seconds of arc */
  
  dms = dms + 0.5/3600.;

  /* Get the whole part in degrees or hours, as needed */  

  d = (int) dms;
  
  /* Could test here for overflow at 24h or 360d          */
  /*   but we'll assume that's been done before the call. */
    
  ms = dms - (double) d;
  ms = ms*60.;
  m = (int) ms;
  s = (int) 60.*(ms - (double) m);
  if (negflag)
  {
    sprintf(outstr,"-%02d:%02d:%02d",d,m,s);
  }
  else
  {
    sprintf(outstr,"%02d:%02d:%02d",d,m,s);      
  }
}


/* XPA client utilities */


void tel_send_telescope(void)
{
  return;
}


void tel_receive_telescope(void)
{  
  #define NXPA 10
  int i, j, nresponses;
  size_t buffer_lengths[NXPA]={0};
  char *data_buffer[NXPA]; 
  char *server_names[NXPA];
  char *server_messages[NXPA];
  double tmpra, tmpdec;
  double tmpra2, tmpdec2;
  
  j=0;
  nresponses = XPAGet(NULL, "tel", "telescope", NULL, data_buffer, 
    buffer_lengths, server_names, server_messages, NXPA);
  
  for(i=0; i < nresponses; i++)
  {
    if( server_messages[i] == NULL )
    {          
      j = sscanf(data_buffer[i],"%lf %lf %lf %lf", &tmpra2, &tmpdec2, &tmpra, &tmpdec);
      if (j == 4)  
      {
        telra2 = tmpra2;
        teldec2 = tmpdec2;
        telra = tmpra;
        teldec = tmpdec;    
      }        
    }
    else
    {
    
      /* Process error messages */
    
      fprintf(stderr,"%s (%s)\n", server_messages[i], server_names[i]);
    
    }
    if( server_names[i] )
    {  
      free( server_names[i] );
    }
    if( server_messages[i] )
    { 
     free( server_messages[i] );  
    }
  }
  return;
}


void tel_send_target(void)
{
  #define NXPA 10
  int i, nresponses;
  size_t buffer_length = 0;
  char *data_buffer = NULL;
  char *server_names[NXPA];
  char *server_messages[NXPA];
  char telmsg[64];
  
  sprintf(telmsg, "target %lf %lf", targetra2, targetdec2);
  
  nresponses = XPASet(NULL, "tel", telmsg, NULL, data_buffer,
    buffer_length, server_names, server_messages, NXPA);
  
  for(i=0; i < nresponses; i++)
  {
    if( server_messages[i] != NULL )
    {
    
      /* Process error messages */
    
      printf("%s (%s)\n", server_messages[i], server_names[i]);
    
    }
    if( server_names[i] )
    {  
      free(server_names[i]);
    }
    if( server_messages[i] )
    { 
     free(server_messages[i]);  
    }
  }
}


void tel_receive_target(void)
{  
  #define NXPA 10
  int i, j, nresponses;
  size_t buffer_lengths[NXPA]={0};
  char *data_buffer[NXPA]; 
  char *server_names[NXPA];
  char *server_messages[NXPA];
  double tmpra, tmpdec;
  double tmpra2, tmpdec2;
  
  j=0;
  nresponses = XPAGet(NULL, "tel", "target", NULL, data_buffer, 
    buffer_lengths, server_names, server_messages, NXPA);
  
  for(i=0; i < nresponses; i++)
  {
    if( server_messages[i] == NULL )
    {          
      j = sscanf(data_buffer[i],"%lf %lf %lf %lf", &tmpra2, &tmpdec2, &tmpra, &tmpdec);
      if (j == 4)  
      {
        targetra2 = tmpra2;
        targetdec2 = tmpdec2;
        targetra = tmpra;
        targetdec = tmpdec;    
      }        
    }
    else
    {
    
      /* Process error messages */
    
      fprintf(stderr,"%s (%s)\n", server_messages[i], server_names[i]);
    
    }
    if( server_names[i] )
    {  
      free( server_names[i] );
    }
    if( server_messages[i] )
    { 
     free( server_messages[i] );  
    }
  }
  return;
}


void tel_send_status(void)
{
  #define NXPA 10
  int i, nresponses;
  size_t buffer_length = 0;
  char *data_buffer = NULL;
  char *server_names[NXPA];
  char *server_messages[NXPA];
  char telmsg[64];

  if ( telflag == STATUSON )
  {
    sprintf(telmsg, "status on");
  }
  else if ( telflag == STATUSOFF )
  {
    sprintf(telmsg, "status off");    
  }
    
  nresponses = XPASet(NULL, "tel", telmsg, NULL, data_buffer,
    buffer_length, server_names, server_messages, NXPA);
  
  for(i=0; i < nresponses; i++)
  {
    if( server_messages[i] != NULL )
    {
    
      /* Process error messages */
    
      printf("%s (%s)\n", server_messages[i], server_names[i]);
    
    }
    if( server_names[i] )
    {  
      free(server_names[i]);
    }
    if( server_messages[i] )
    { 
     free(server_messages[i]);  
    }
  }
}

void tel_receive_status(void)
{  
  #define NXPA 10
  int i, j, nresponses;
  size_t buffer_lengths[NXPA]={0};
  char *data_buffer[NXPA]; 
  char *server_names[NXPA];
  char *server_messages[NXPA];
  char telmsg[64];
   
  j=0;
  nresponses = XPAGet(NULL, "tel", "status", NULL, data_buffer, 
    buffer_lengths, server_names, server_messages, NXPA);
  
  for(i=0; i < nresponses; i++)
  {
    if( server_messages[i] == NULL )
    {          
      j = sscanf(data_buffer[i],"%s", telmsg);
      if (j == 1)  
      {
        if ( !strcmp(telmsg, "off") )
        {
          telflag = STATUSOFF;
        }    
        else if ( !strcmp(telmsg, "on") )
        {
          telflag = STATUSON;    
        }    
      
      }        
    }
    else
    {
    
      /* Process error messages */
    
      fprintf(stderr,"%s (%s)\n", server_messages[i], server_names[i]);
    
    }
    if( server_names[i] )
    {  
      free( server_names[i] );
    }
    if( server_messages[i] )
    { 
     free( server_messages[i] );  
    }
  }
  return;
}


void tel_send_track(void)
{
  #define NXPA 10
  int i, nresponses;
  size_t buffer_length = 0;
  char *data_buffer = NULL;
  char *server_names[NXPA];
  char *server_messages[NXPA];
  char telmsg[64];

  if ( trackflag == TRUE )
  {
    sprintf(telmsg, "track on");
  }
  else 
  {
    sprintf(telmsg, "track off");
  }
    
  nresponses = XPASet(NULL, "tel", telmsg, NULL, data_buffer,
    buffer_length, server_names, server_messages, NXPA);
  
  for(i=0; i < nresponses; i++)
  {
    if( server_messages[i] != NULL )
    {
    
      /* Process error messages */
    
      printf("%s (%s)\n", server_messages[i], server_names[i]);
    
    }
    if( server_names[i] )
    {  
      free(server_names[i]);
    }
    if( server_messages[i] )
    { 
     free(server_messages[i]);  
    }
  }
}

void tel_receive_track(void)
{  
  #define NXPA 10
  int i, j, nresponses;
  size_t buffer_lengths[NXPA]={0};
  char *data_buffer[NXPA]; 
  char *server_names[NXPA];
  char *server_messages[NXPA];
  char telmsg[64];
   
  j=0;
  nresponses = XPAGet(NULL, "tel", "track", NULL, data_buffer, 
    buffer_lengths, server_names, server_messages, NXPA);
  
  for(i=0; i < nresponses; i++)
  {
    if( server_messages[i] == NULL )
    {          
      j = sscanf(data_buffer[i],"%s", telmsg);
      if (j == 1)  
      {
        if ( !strcmp(telmsg, "on") )
        {
          trackflag = TRUE;
        }    
        else if ( !strcmp(telmsg, "off") )
        {
          trackflag = FALSE;    
        }    
      
      }        
    }
    else
    {
    
      /* Process error messages */
    
      fprintf(stderr,"%s (%s)\n", server_messages[i], server_names[i]);
    
    }
    if( server_names[i] )
    {  
      free( server_names[i] );
    }
    if( server_messages[i] )
    { 
     free( server_messages[i] );  
    }
  }
  return;
}


void tel_send_goto(void)
{
  #define NXPA 10
  int i, nresponses;
  size_t buffer_length = 0;
  char *data_buffer = NULL;
  char *server_names[NXPA];
  char *server_messages[NXPA];
  char telmsg[64];

  if ( gotoflag == TRUE )
  {
    sprintf(telmsg, "goto on");
  }
  else 
  {
    sprintf(telmsg, "goto off");
  }
    
  nresponses = XPASet(NULL, "tel", telmsg, NULL, data_buffer,
    buffer_length, server_names, server_messages, NXPA);
  
  for(i=0; i < nresponses; i++)
  {
    if( server_messages[i] != NULL )
    {
    
      /* Process error messages */
    
      printf("%s (%s)\n", server_messages[i], server_names[i]);
    
    }
    if( server_names[i] )
    {  
      free(server_names[i]);
    }
    if( server_messages[i] )
    { 
     free(server_messages[i]);  
    }
  }
}

void tel_receive_goto(void)
{  
  #define NXPA 10
  int i, j, nresponses;
  size_t buffer_lengths[NXPA]={0};
  char *data_buffer[NXPA]; 
  char *server_names[NXPA];
  char *server_messages[NXPA];
  char telmsg[64];
   
  j=0;
  nresponses = XPAGet(NULL, "tel", "goto", NULL, data_buffer, 
    buffer_lengths, server_names, server_messages, NXPA);
  
  for(i=0; i < nresponses; i++)
  {
    if( server_messages[i] == NULL )
    {          
      j = sscanf(data_buffer[i],"%s", telmsg);
      if (j == 1)  
      {
        if ( !strcmp(telmsg, "on") )
        {
          gotoflag = TRUE;
        }    
        else if ( !strcmp(telmsg, "off") )
        {
          gotoflag = FALSE;    
        }    
      
      }        
    }
    else
    {
    
      /* Process error messages */
    
      fprintf(stderr,"%s (%s)\n", server_messages[i], server_names[i]);
    
    }
    if( server_names[i] )
    {  
      free( server_names[i] );
    }
    if( server_messages[i] )
    { 
     free( server_messages[i] );  
    }
  }
  return;
}


void tel_send_guide(void)
{
  #define NXPA 10
  int i, nresponses;
  size_t buffer_length = 0;
  char *data_buffer = NULL;
  char *server_names[NXPA];
  char *server_messages[NXPA];
  char telmsg[64];

  if ( guideflag == TRUE )
  {
    sprintf(telmsg, "guide on");
  }
  else 
  {
    sprintf(telmsg, "guide off");
  }
    
  nresponses = XPASet(NULL, "tel", telmsg, NULL, data_buffer,
    buffer_length, server_names, server_messages, NXPA);
  
  for(i=0; i < nresponses; i++)
  {
    if( server_messages[i] != NULL )
    {
    
      /* Process error messages */
    
      printf("%s (%s)\n", server_messages[i], server_names[i]);
    
    }
    if( server_names[i] )
    {  
      free(server_names[i]);
    }
    if( server_messages[i] )
    { 
     free(server_messages[i]);  
    }
  }
}

void tel_receive_guide(void)
{  
  #define NXPA 10
  int i, j, nresponses;
  size_t buffer_lengths[NXPA]={0};
  char *data_buffer[NXPA]; 
  char *server_names[NXPA];
  char *server_messages[NXPA];
  char telmsg[64];
   
  j=0;
  nresponses = XPAGet(NULL, "tel", "guide", NULL, data_buffer, 
    buffer_lengths, server_names, server_messages, NXPA);
  
  for(i=0; i < nresponses; i++)
  {
    if( server_messages[i] == NULL )
    {          
      j = sscanf(data_buffer[i],"%s", telmsg);
      if (j == 1)  
      {
        if ( !strcmp(telmsg, "on") )
        {
          guideflag = TRUE;
        }    
        else if ( !strcmp(telmsg, "off") )
        {
          guideflag = FALSE;    
        }    
      
      }        
    }
    else
    {
    
      /* Process error messages */
    
      fprintf(stderr,"%s (%s)\n", server_messages[i], server_names[i]);
    
    }
    if( server_names[i] )
    {  
      free( server_names[i] );
    }
    if( server_messages[i] )
    { 
     free( server_messages[i] );  
    }
  }
  return;
}


void tel_send_speed(void)
{
  #define NXPA 10
  int i, nresponses;
  size_t buffer_length = 0;
  char *data_buffer = NULL;
  char *server_names[NXPA];
  char *server_messages[NXPA];
  char telmsg[64];

  if ( telspd == SLEW)
  {
    sprintf(telmsg, "speed slew");
  }
  else if ( telspd == FIND)
  {
    sprintf(telmsg, "speed find");
  }
  else if ( telspd == CENTER)
  {
    sprintf(telmsg, "speed center");
  }
  else if ( telspd == GUIDE)
  {
    sprintf(telmsg, "speed guide");
  }
  else
  {
    return;
  } 
    
  nresponses = XPASet(NULL, "tel", telmsg, NULL, data_buffer,
    buffer_length, server_names, server_messages, NXPA);
  
  for(i=0; i < nresponses; i++)
  {
    if( server_messages[i] != NULL )
    {
    
      /* Process error messages */
    
      printf("%s (%s)\n", server_messages[i], server_names[i]);
    
    }
    if( server_names[i] )
    {  
      free(server_names[i]);
    }
    if( server_messages[i] )
    { 
     free(server_messages[i]);  
    }
  }
}

void tel_receive_speed(void)
{  
  #define NXPA 10
  int i, j, nresponses;
  size_t buffer_lengths[NXPA]={0};
  char *data_buffer[NXPA]; 
  char *server_names[NXPA];
  char *server_messages[NXPA];
  char telmsg[64];
   
  j=0;
  nresponses = XPAGet(NULL, "tel", "speed", NULL, data_buffer, 
    buffer_lengths, server_names, server_messages, NXPA);
  
  for(i=0; i < nresponses; i++)
  {
    if( server_messages[i] == NULL )
    {          
      j = sscanf(data_buffer[i],"%s", telmsg);
      if (j == 1)  
      {
        if ( !strcmp(telmsg, "slew") )
        {
          telspd = SLEW;
        }    
        else if ( !strcmp(telmsg, "find") )
        {
          telspd = FIND; 
        }    
        else if ( !strcmp(telmsg, "center") )
        {
          telspd = CENTER; 
        }    
        else if ( !strcmp(telmsg, "guide") )
        {
          telspd = GUIDE; 
        }
        else
        {
          return;
        }      
      }        
    }
    else
    {
    
      /* Process error messages */
    
      fprintf(stderr,"%s (%s)\n", server_messages[i], server_names[i]);
    
    }
    if( server_names[i] )
    {  
      free( server_names[i] );
    }
    if( server_messages[i] )
    { 
     free( server_messages[i] );  
    }
  }
  return;
}

void tel_send_move(void)
{
  #define NXPA 10
  int i, nresponses;
  size_t buffer_length = 0;
  char *data_buffer = NULL;
  char *server_names[NXPA];
  char *server_messages[NXPA];
  char telmsg[64];

  if ( teldir == NORTH )
  {
    sprintf(telmsg, "move north %lf", movetime);
  }
  else if ( teldir == SOUTH )
  {
    sprintf(telmsg, "move south %lf", movetime);
  }
  else if ( teldir == EAST )
  {
    sprintf(telmsg, "move east %lf", movetime);
  }
  else if ( teldir == WEST )
  {
    sprintf(telmsg, "move west %lf", movetime);
  }
  else
  {
    return;
  }  
      
  nresponses = XPASet(NULL, "tel", telmsg, NULL, data_buffer,
    buffer_length, server_names, server_messages, NXPA);
  
  for(i=0; i < nresponses; i++)
  {
    if( server_messages[i] != NULL )
    {
    
      /* Process error messages */
    
      printf("%s (%s)\n", server_messages[i], server_names[i]);
    
    }
    if( server_names[i] )
    {  
      free(server_names[i]);
    }
    if( server_messages[i] )
    { 
     free(server_messages[i]);  
    }
  }
}

void tel_receive_move(void)
{  
  return;
}


void tel_send_refract(void)
{
  #define NXPA 10
  int i, nresponses;
  size_t buffer_length = 0;
  char *data_buffer = NULL;
  char *server_names[NXPA];
  char *server_messages[NXPA];
  char telmsg[64];

  if ( pmodel == (pmodel | REFRACT) )
  {
    sprintf(telmsg, "refract on");
  }
  else 
  {
    sprintf(telmsg, "refract off");
  }
    
  nresponses = XPASet(NULL, "tel", telmsg, NULL, data_buffer,
    buffer_length, server_names, server_messages, NXPA);
  
  for(i=0; i < nresponses; i++)
  {
    if( server_messages[i] != NULL )
    {
    
      /* Process error messages */
    
      printf("%s (%s)\n", server_messages[i], server_names[i]);
    
    }
    if( server_names[i] )
    {  
      free(server_names[i]);
    }
    if( server_messages[i] )
    { 
     free(server_messages[i]);  
    }
  }
}

void tel_receive_refract(void)
{  
  #define NXPA 10
  int i, j, nresponses;
  size_t buffer_lengths[NXPA]={0};
  char *data_buffer[NXPA]; 
  char *server_names[NXPA];
  char *server_messages[NXPA];
  char telmsg[64];
   
  j=0;
  nresponses = XPAGet(NULL, "tel", "refract", NULL, data_buffer, 
    buffer_lengths, server_names, server_messages, NXPA);
  
  for(i=0; i < nresponses; i++)
  {
    if( server_messages[i] == NULL )
    {          
      j = sscanf(data_buffer[i],"%s", telmsg);
      if (j == 1)  
      {
        if ( !strcmp(telmsg, "on") )
        {
          pmodel = (pmodel | REFRACT);
        }    
        else if ( !strcmp(telmsg, "off") )
        {
          pmodel = (pmodel & ~(REFRACT));  
        }    
      
      }        
    }
    else
    {
    
      /* Process error messages */
    
      fprintf(stderr,"%s (%s)\n", server_messages[i], server_names[i]);
    
    }
    if( server_names[i] )
    {  
      free( server_names[i] );
    }
    if( server_messages[i] )
    { 
     free( server_messages[i] );  
    }
  }
  return;
}


void tel_send_dynamic(void)
{
  #define NXPA 10
  int i, nresponses;
  size_t buffer_length = 0;
  char *data_buffer = NULL;
  char *server_names[NXPA];
  char *server_messages[NXPA];
  char telmsg[64];

  if ( pmodel == (pmodel | DYNAMIC) )
  {
    sprintf(telmsg, "dynamic on %lf %lf", modelha1, modeldec1);
  }
  else 
  {
    sprintf(telmsg, "dynamic off %lf %lf", modelha1, modeldec1);
  }
    
  nresponses = XPASet(NULL, "tel", telmsg, NULL, data_buffer,
    buffer_length, server_names, server_messages, NXPA);
  
  for(i=0; i < nresponses; i++)
  {
    if( server_messages[i] != NULL )
    {
    
      /* Process error messages */
    
      printf("%s (%s)\n", server_messages[i], server_names[i]);
    
    }
    if( server_names[i] )
    {  
      free(server_names[i]);
    }
    if( server_messages[i] )
    { 
     free(server_messages[i]);  
    }
  }
}

void tel_receive_dynamic(void)
{  
  #define NXPA 10
  int i, j, nresponses;
  size_t buffer_lengths[NXPA]={0};
  char *data_buffer[NXPA]; 
  char *server_names[NXPA];
  char *server_messages[NXPA];
  char telmsg[64];
  double tmpha1, tmpdec1;
   
  j=0;
  nresponses = XPAGet(NULL, "tel", "dynamic", NULL, data_buffer, 
    buffer_lengths, server_names, server_messages, NXPA);
  
  for(i=0; i < nresponses; i++)
  {
    if( server_messages[i] == NULL )
    {          
      j = sscanf(data_buffer[i],"%s %lf %lf", telmsg, &tmpha1, &tmpdec1);
      if (j == 3)  
      {
        if ( !strcmp(telmsg, "on") )
        {
          pmodel = (pmodel | DYNAMIC);
          modelha1 = tmpha1;
          modeldec1 = tmpdec1;
        }    
        else if ( !strcmp(telmsg, "off") )
        {
          pmodel = (pmodel & ~(DYNAMIC));
          modelha1 = tmpha1;
          modeldec1 = tmpdec1;            
        }    
      
      }        
    }
    else
    {
    
      /* Process error messages */
    
      fprintf(stderr,"%s (%s)\n", server_messages[i], server_names[i]);
    
    }
    if( server_names[i] )
    {  
      free( server_names[i] );
    }
    if( server_messages[i] )
    { 
     free( server_messages[i] );  
    }
  }
  return;
}


void tel_send_offset(void)
{
  #define NXPA 10
  int i, nresponses;
  size_t buffer_length = 0;
  char *data_buffer = NULL;
  char *server_names[NXPA];
  char *server_messages[NXPA];
  char telmsg[64];

  if ( pmodel == (pmodel | OFFSET) )
  {
    sprintf(telmsg, "offset on");
  }
  else 
  {
    sprintf(telmsg, "offset off");
  }
    
  nresponses = XPASet(NULL, "tel", telmsg, NULL, data_buffer,
    buffer_length, server_names, server_messages, NXPA);
  
  for(i=0; i < nresponses; i++)
  {
    if( server_messages[i] != NULL )
    {
    
      /* Process error messages */
    
      printf("%s (%s)\n", server_messages[i], server_names[i]);
    
    }
    if( server_names[i] )
    {  
      free(server_names[i]);
    }
    if( server_messages[i] )
    { 
     free(server_messages[i]);  
    }
  }
}

void tel_send_offset_target(void)
{
  #define NXPA 10
  int i, nresponses;
  size_t buffer_length = 0;
  char *data_buffer = NULL;
  char *server_names[NXPA];
  char *server_messages[NXPA];
  char telmsg[64];

  if ( pmodel == (pmodel | OFFSET) )
  {
    sprintf(telmsg, "offset target");
  }
  else 
  {
    sprintf(telmsg, "offset off");
  }
    
  nresponses = XPASet(NULL, "tel", telmsg, NULL, data_buffer,
    buffer_length, server_names, server_messages, NXPA);
  
  for(i=0; i < nresponses; i++)
  {
    if( server_messages[i] != NULL )
    {
    
      /* Process error messages */
    
      printf("%s (%s)\n", server_messages[i], server_names[i]);
    
    }
    if( server_names[i] )
    {  
      free(server_names[i]);
    }
    if( server_messages[i] )
    { 
     free(server_messages[i]);  
    }
  }
}

void tel_send_offset_clear(void)
{
  #define NXPA 10
  int i, nresponses;
  size_t buffer_length = 0;
  char *data_buffer = NULL;
  char *server_names[NXPA];
  char *server_messages[NXPA];
  char telmsg[64];

  if ( pmodel == (pmodel | OFFSET) )
  {
    sprintf(telmsg, "offset clear");
  }
  else 
  {
    sprintf(telmsg, "offset off");
  }
    
  nresponses = XPASet(NULL, "tel", telmsg, NULL, data_buffer,
    buffer_length, server_names, server_messages, NXPA);
  
  for(i=0; i < nresponses; i++)
  {
    if( server_messages[i] != NULL )
    {
    
      /* Process error messages */
    
      printf("%s (%s)\n", server_messages[i], server_names[i]);
    
    }
    if( server_names[i] )
    {  
      free(server_names[i]);
    }
    if( server_messages[i] )
    { 
     free(server_messages[i]);  
    }
  }
}

void tel_receive_offset(void)
{  
  #define NXPA 10
  int i, j, nresponses;
  size_t buffer_lengths[NXPA]={0};
  char *data_buffer[NXPA]; 
  char *server_names[NXPA];
  char *server_messages[NXPA];
  char telmsg[64];
  double tmpoffsetha, tmpoffsetdec;
   
  j=0;
  nresponses = XPAGet(NULL, "tel", "offset", NULL, data_buffer, 
    buffer_lengths, server_names, server_messages, NXPA);
  
  for(i=0; i < nresponses; i++)
  {
    if( server_messages[i] == NULL )
    {          
      j = sscanf(data_buffer[i],"%s %lf %lf", telmsg, &tmpoffsetha, &tmpoffsetdec);
      if ((j == 3) && (strcmp(telmsg,"on") == 0))
      {
        if ((fabs(tmpoffsetha) <= 15.0) && (fabs(tmpoffsetdec) <= 1.0))
        {
          offsetdec = tmpoffsetdec;
          offsetha = tmpoffsetha;
          pmodel = (pmodel | OFFSET);
        }      
      }
      else if ((j == 3) && (strcmp(telmsg,"off") == 0))
      {
        if ((fabs(tmpoffsetha) <= 15.0) && (fabs(tmpoffsetdec) <= 1.0))
        {
          offsetdec = tmpoffsetdec;
          offsetha = tmpoffsetha;
          pmodel = (pmodel & ~(OFFSET));
        }      
      }
      else if ((j == 1) && (strcmp(telmsg,"on") == 0))
      {
        pmodel = (pmodel | OFFSET);
      }
      else if ((j == 1) && (strcmp(telmsg,"off") == 0))
      {
        pmodel = (pmodel & ~(OFFSET));      
      }
      else if ((j == 1) && (strcmp(telmsg,"target") == 0))
      {
        tmpoffsetdec = targetdec - teldec; 
        tmpoffsetha =  Map12(telra - targetra);  
          
        if ((fabs(tmpoffsetha) <= 15.0) && (fabs(tmpoffsetdec) <= 1.0))
        {
          offsetdec = tmpoffsetdec;
          offsetha = tmpoffsetha;
          pmodel = (pmodel | OFFSET);
        }      
        
      }
      else if ((j == 1) && (strcmp(telmsg,"clear") == 0))
      {
        pmodel = (pmodel & ~(OFFSET));
        offsetdec = 0.; 
        offsetha = 0.;                 
      }        
    }
    else
    {
    
      /* Process error messages */
    
      fprintf(stderr,"%s (%s)\n", server_messages[i], server_names[i]);
    
    }
    if( server_names[i] )
    {  
      free( server_names[i] );
    }
    if( server_messages[i] )
    { 
     free( server_messages[i] );  
    }
  }
  return;
}

void tel_send_polar(void)
{
  #define NXPA 10
  int i, nresponses;
  size_t buffer_length = 0;
  char *data_buffer = NULL;
  char *server_names[NXPA];
  char *server_messages[NXPA];
  char telmsg[64];

  if ( pmodel == (pmodel | POLAR) )
  {
    sprintf(telmsg, "polar on %lf %lf", polaraz, polaralt);
  }
  else 
  {
    sprintf(telmsg, "polar off %lf %lf", polaraz, polaralt);
  }
    
  nresponses = XPASet(NULL, "tel", telmsg, NULL, data_buffer,
    buffer_length, server_names, server_messages, NXPA);
  
  for(i=0; i < nresponses; i++)
  {
    if( server_messages[i] != NULL )
    {
    
      /* Process error messages */
    
      printf("%s (%s)\n", server_messages[i], server_names[i]);
    
    }
    if( server_names[i] )
    {  
      free(server_names[i]);
    }
    if( server_messages[i] )
    { 
     free(server_messages[i]);  
    }
  }
}

void tel_receive_polar(void)
{  
  #define NXPA 10
  int i, j, nresponses;
  size_t buffer_lengths[NXPA]={0};
  char *data_buffer[NXPA]; 
  char *server_names[NXPA];
  char *server_messages[NXPA];
  char telmsg[64];
  double tmpaz, tmpalt;
   
  j=0;
  nresponses = XPAGet(NULL, "tel", "polar", NULL, data_buffer, 
    buffer_lengths, server_names, server_messages, NXPA);
  
  for(i=0; i < nresponses; i++)
  {
    if( server_messages[i] == NULL )
    {          
      j = sscanf(data_buffer[i],"%s %lf %lf", telmsg, &tmpaz, &tmpalt);
      if ((j == 3) && (strcmp(telmsg, "on") == 0))
      {
        pmodel = (pmodel | POLAR);
        polaraz = tmpaz;
        polaralt = tmpalt;
      }
      else if ((j == 3) && (strcmp(telmsg, "off") == 0 ))
      {
        pmodel = (pmodel & ~(POLAR));
        polaraz = tmpaz;
        polaralt = tmpalt;
      }
      else if ((j == 1) && (strcmp(telmsg, "on") == 0 ))
      {
        pmodel = (pmodel | POLAR);
      }
      else if ((j == 1) && (strcmp(telmsg, "off") == 0 ))
      {
        pmodel = (pmodel & ~(POLAR));      
      }
      else if ((j == 1) && (strcmp(telmsg, "clear") == 0 ))
      {
        pmodel = (pmodel & ~(POLAR));
        polaralt = 0.;
        polaraz = 0.;      
      }       
    }
    else
    {
    
      /* Process error messages */
    
      fprintf(stderr,"%s (%s)\n", server_messages[i], server_names[i]);
    
    }
    if( server_names[i] )
    {  
      free( server_names[i] );
    }
    if( server_messages[i] )
    { 
     free( server_messages[i] );  
    }
  }
  return;
}

void tel_send_park(void)
{
  #define NXPA 10
  int i, nresponses;
  size_t buffer_length = 0;
  char *data_buffer = NULL;
  char *server_names[NXPA];
  char *server_messages[NXPA];
  char telmsg[64];

  if ( parkflag == TRUE )
  {
    sprintf(telmsg, "park on");
  }
  else 
  {
    sprintf(telmsg, "park off");
  }
    
  nresponses = XPASet(NULL, "tel", telmsg, NULL, data_buffer,
    buffer_length, server_names, server_messages, NXPA);
  
  for(i=0; i < nresponses; i++)
  {
    if( server_messages[i] != NULL )
    {
    
      /* Process error messages */
    
      printf("%s (%s)\n", server_messages[i], server_names[i]);
    
    }
    if( server_names[i] )
    {  
      free(server_names[i]);
    }
    if( server_messages[i] )
    { 
     free(server_messages[i]);  
    }
  }
}

void tel_receive_park(void)
{  
  #define NXPA 10
  int i, j, nresponses;
  size_t buffer_lengths[NXPA]={0};
  char *data_buffer[NXPA]; 
  char *server_names[NXPA];
  char *server_messages[NXPA];
  char telmsg[64];
   
  j=0;
  nresponses = XPAGet(NULL, "tel", "park", NULL, data_buffer, 
    buffer_lengths, server_names, server_messages, NXPA);
  
  for(i=0; i < nresponses; i++)
  {
    if( server_messages[i] == NULL )
    {          
      j = sscanf(data_buffer[i],"%s", telmsg);
      if (j == 1)  
      {
        if ( !strcmp(telmsg, "on") )
        {
          parkflag = TRUE;
        }    
        else if ( !strcmp(telmsg, "off") )
        {
          parkflag = FALSE;    
        }    
      
      }        
    }
    else
    {
    
      /* Process error messages */
    
      fprintf(stderr,"%s (%s)\n", server_messages[i], server_names[i]);
    
    }
    if( server_names[i] )
    {  
      free( server_names[i] );
    }
    if( server_messages[i] )
    { 
     free( server_messages[i] );  
    }
  }
  return;
}

void tel_send_home(void)
{
  #define NXPA 10
  int i, nresponses;
  size_t buffer_length = 0;
  char *data_buffer = NULL;
  char *server_names[NXPA];
  char *server_messages[NXPA];
  char telmsg[64];

  if ( homeflag == TRUE )
  {
    sprintf(telmsg, "home on");
  }
  else 
  {
    sprintf(telmsg, "home off");
  }
    
  nresponses = XPASet(NULL, "tel", telmsg, NULL, data_buffer,
    buffer_length, server_names, server_messages, NXPA);
  
  for(i=0; i < nresponses; i++)
  {
    if( server_messages[i] != NULL )
    {
    
      /* Process error messages */
    
      printf("%s (%s)\n", server_messages[i], server_names[i]);
    
    }
    if( server_names[i] )
    {  
      free(server_names[i]);
    }
    if( server_messages[i] )
    { 
     free(server_messages[i]);  
    }
  }
}

void tel_receive_home(void)
{  
  #define NXPA 10
  int i, j, nresponses;
  size_t buffer_lengths[NXPA]={0};
  char *data_buffer[NXPA]; 
  char *server_names[NXPA];
  char *server_messages[NXPA];
  char telmsg[64];
   
  j=0;
  nresponses = XPAGet(NULL, "tel", "home", NULL, data_buffer, 
    buffer_lengths, server_names, server_messages, NXPA);
  
  for(i=0; i < nresponses; i++)
  {
    if( server_messages[i] == NULL )
    {          
      j = sscanf(data_buffer[i],"%s", telmsg);
      if (j == 1)  
      {
        if ( !strcmp(telmsg, "on") )
        {
          homeflag = TRUE;
        }    
        else if ( !strcmp(telmsg, "off") )
        {
          homeflag = FALSE;    
        }    
      
      }        
    }
    else
    {
    
      /* Process error messages */
    
      fprintf(stderr,"%s (%s)\n", server_messages[i], server_names[i]);
    
    }
    if( server_names[i] )
    {  
      free( server_names[i] );
    }
    if( server_messages[i] )
    { 
     free( server_messages[i] );  
    }
  }
  return;
}

